# pylint: disable=too-many-lines
from io import StringIO
import sys
from time import sleep
from unittest import mock, TestCase
import xml.dom.minidom
import xml.etree.ElementTree as ET

from pcs_test.tools.xml import dom_get_child_elements
from pcs_test.tools.misc import get_test_resource as rc

from pcs import utils

# pylint: disable=line-too-long
# pylint: disable=invalid-name

cib_with_nodes = rc("cib-empty-withnodes.xml")
empty_cib = rc("cib-empty.xml")

TestCase.maxDiff = None


class UtilsTest(TestCase):
    # pylint: disable=too-many-public-methods
    @staticmethod
    def get_cib_empty():
        return xml.dom.minidom.parse(empty_cib)

    @staticmethod
    def get_cib_with_nodes_minidom():
        return xml.dom.minidom.parse(cib_with_nodes)

    def get_cib_resources(self):
        cib_dom = self.get_cib_empty()
        new_resources = xml.dom.minidom.parseString(
            """
            <resources>
                  <primitive id="myResource"
                        class="ocf" provider="heartbeat" type="Dummy">
                  </primitive>
                  <clone id="myClone">
                      <primitive id="myClonedResource"
                          class="ocf" provider="heartbeat" type="Dummy">
                      </primitive>
                  </clone>
                  <clone id="myUniqueClone">
                      <primitive id="myUniqueClonedResource"
                          class="ocf" provider="heartbeat" type="Dummy">
                      </primitive>
                      <meta-attributes>
                        <nvpair name="globally-unique" value="true" />
                      </meta-attributes>
                  </clone>
                  <master id="myMaster">
                      <primitive id="myMasteredResource"
                            class="ocf" provider="heartbeat" type="Dummy">
                      </primitive>
                  </master>
                  <group id="myGroup">
                      <primitive id="myGroupedResource"
                            class="ocf" provider="heartbeat" type="Dummy">
                      </primitive>
                  </group>
                  <clone id="myGroupClone">
                      <group id="myClonedGroup">
                          <primitive id="myClonedGroupedResource"
                                class="ocf" provider="heartbeat" type="Dummy">
                          </primitive>
                      </group>
                  </clone>
                  <master id="myGroupMaster">
                      <group id="myMasteredGroup">
                          <primitive id="myMasteredGroupedResource"
                                class="ocf" provider="heartbeat" type="Dummy">
                          </primitive>
                      </group>
                  </master>
                  <bundle id="myBundle">
                      <primitive id="myBundledResource"
                          class="ocf" provider="heartbeat" type="Dummy" />
                  </bundle>
                  <bundle id="myEmptyBundle"/>
            </resources>
        """
        ).documentElement
        resources = cib_dom.getElementsByTagName("resources")[0]
        resources.parentNode.replaceChild(new_resources, resources)
        return cib_dom

    def testDomGetResources(self):
        # pylint: disable=too-many-statements
        def test_dom_get(method, dom, ok_ids, bad_ids):
            for element_id in ok_ids:
                self.assert_element_id(method(dom, element_id), element_id)
            for element_id in bad_ids:
                self.assertFalse(method(dom, element_id))

        cib_dom = self.get_cib_empty()
        self.assertFalse(utils.dom_get_resource(cib_dom, "myResource"))
        self.assertFalse(
            utils.dom_get_resource_clone(cib_dom, "myClonedResource")
        )
        self.assertFalse(
            utils.dom_get_resource_masterslave(cib_dom, "myMasteredResource")
        )
        self.assertFalse(utils.dom_get_group(cib_dom, "myGroup"))
        self.assertFalse(utils.dom_get_group_clone(cib_dom, "myClonedGroup"))
        self.assertFalse(
            utils.dom_get_group_masterslave(cib_dom, "myMasteredGroup")
        )
        self.assertFalse(utils.dom_get_clone(cib_dom, "myClone"))
        self.assertFalse(utils.dom_get_master(cib_dom, "myMaster"))
        self.assertFalse(utils.dom_get_clone_ms_resource(cib_dom, "myClone"))
        self.assertFalse(utils.dom_get_clone_ms_resource(cib_dom, "myMaster"))
        self.assertFalse(
            utils.dom_get_resource_clone_ms_parent(cib_dom, "myClonedResource")
        )
        self.assertFalse(
            utils.dom_get_resource_clone_ms_parent(
                cib_dom, "myMasteredResource"
            )
        )
        self.assertIsNone(utils.dom_get_bundle(cib_dom, "myResource"))
        self.assertIsNone(utils.dom_get_bundle(cib_dom, "notExisting"))
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myBundledResource")
        )

        cib_dom = self.get_cib_resources()
        all_ids = set(
            [
                "none",
                "myResource",
                "myClone",
                "myClonedResource",
                "myUniqueClone",
                "myUniqueClonedResource",
                "myMaster",
                "myMasteredResource",
                "myGroup",
                "myGroupedResource",
                "myGroupClone",
                "myClonedGroup",
                "myClonedGroupedResource",
                "myGroupMaster",
                "myMasteredGroup",
                "myMasteredGroupedResource",
                "myBundledResource",
                "myBundle",
                "myEmptyBundle",
            ]
        )

        resource_ids = set(
            [
                "myResource",
                "myClonedResource",
                "myUniqueClonedResource",
                "myGroupedResource",
                "myMasteredResource",
                "myClonedGroupedResource",
                "myMasteredGroupedResource",
                "myBundledResource",
            ]
        )
        test_dom_get(
            utils.dom_get_resource,
            cib_dom,
            resource_ids,
            all_ids - resource_ids,
        )

        cloned_ids = set(
            [
                "myClonedResource",
                "myUniqueClonedResource",
                "myClonedGroupedResource",
            ]
        )
        test_dom_get(
            utils.dom_get_resource_clone,
            cib_dom,
            cloned_ids,
            all_ids - cloned_ids,
        )

        mastered_ids = set(["myMasteredResource", "myMasteredGroupedResource"])
        test_dom_get(
            utils.dom_get_resource_masterslave,
            cib_dom,
            mastered_ids,
            all_ids - mastered_ids,
        )

        group_ids = set(["myGroup", "myClonedGroup", "myMasteredGroup"])
        test_dom_get(
            utils.dom_get_group, cib_dom, group_ids, all_ids - group_ids
        )

        cloned_group_ids = set(["myClonedGroup"])
        test_dom_get(
            utils.dom_get_group_clone,
            cib_dom,
            cloned_group_ids,
            all_ids - cloned_group_ids,
        )

        clone_ids = set(["myClone", "myUniqueClone", "myGroupClone"])
        test_dom_get(
            utils.dom_get_clone, cib_dom, clone_ids, all_ids - clone_ids
        )

        mastered_group_ids = set(["myMasteredGroup"])
        test_dom_get(
            utils.dom_get_group_masterslave,
            cib_dom,
            mastered_group_ids,
            all_ids - mastered_group_ids,
        )

        master_ids = set(["myMaster", "myGroupMaster"])
        test_dom_get(
            utils.dom_get_master, cib_dom, master_ids, all_ids - master_ids
        )

        bundle_ids = set(["myBundle", "myEmptyBundle"])
        test_dom_get(
            utils.dom_get_bundle, cib_dom, bundle_ids, all_ids - bundle_ids
        )

        self.assert_element_id(
            utils.dom_get_clone_ms_resource(cib_dom, "myClone"),
            "myClonedResource",
        )
        self.assert_element_id(
            utils.dom_get_clone_ms_resource(cib_dom, "myGroupClone"),
            "myClonedGroup",
        )
        self.assert_element_id(
            utils.dom_get_clone_ms_resource(cib_dom, "myMaster"),
            "myMasteredResource",
        )
        self.assert_element_id(
            utils.dom_get_clone_ms_resource(cib_dom, "myGroupMaster"),
            "myMasteredGroup",
        )

        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(cib_dom, "myClonedResource"),
            "myClone",
        )
        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(cib_dom, "myClonedGroup"),
            "myGroupClone",
        )
        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(
                cib_dom, "myClonedGroupedResource"
            ),
            "myGroupClone",
        )
        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(
                cib_dom, "myMasteredResource"
            ),
            "myMaster",
        )
        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(cib_dom, "myMasteredGroup"),
            "myGroupMaster",
        )
        self.assert_element_id(
            utils.dom_get_resource_clone_ms_parent(
                cib_dom, "myMasteredGroupedResource"
            ),
            "myGroupMaster",
        )
        self.assertEqual(
            None, utils.dom_get_resource_clone_ms_parent(cib_dom, "myResource")
        )
        self.assertEqual(
            None, utils.dom_get_resource_clone_ms_parent(cib_dom, "myGroup")
        )
        self.assertEqual(
            None,
            utils.dom_get_resource_clone_ms_parent(
                cib_dom, "myGroupedResource"
            ),
        )

        self.assertIsNone(
            utils.dom_get_resource_bundle(
                utils.dom_get_bundle(cib_dom, "myEmptyBundle")
            )
        )
        self.assert_element_id(
            utils.dom_get_resource_bundle(
                utils.dom_get_bundle(cib_dom, "myBundle")
            ),
            "myBundledResource",
            "primitive",
        )

        self.assert_element_id(
            utils.dom_get_resource_bundle_parent(cib_dom, "myBundledResource"),
            "myBundle",
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myResource")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myClone")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myClonedResource")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myMaster")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myMasteredGroup")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myGroup")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myGroupedResource")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myGroupClone")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(cib_dom, "myClonedGroup")
        )
        self.assertIsNone(
            utils.dom_get_resource_bundle_parent(
                cib_dom, "myClonedGroupedResource"
            )
        )

    def testDomGetResourceRemoteNodeName(self):
        dom = self.get_cib_empty()
        new_resources = xml.dom.minidom.parseString(
            """
            <resources>
                <primitive id="dummy1"
                        class="ocf" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="vm-guest1" provider="heartbeat"
                        type="VirtualDomain">
                    <instance_attributes id="vm-guest1-instance_attributes">
                        <nvpair id="vm-guest1-instance_attributes-hypervisor"
                            name="hypervisor" value="qemu:///system"/>
                        <nvpair id="vm-guest1-instance_attributes-config"
                            name="config" value="/root/guest1.xml"/>
                    </instance_attributes>
                    <meta_attributes id="vm-guest1-meta_attributes">
                        <nvpair id="vm-guest1-meta_attributes-remote-node"
                            name="remote-node" value="guest1"/>
                    </meta_attributes>
                </primitive>
                <primitive id="dummy2"
                        class="ocf" provider="heartbeat" type="Dummy">
                    <instance_attributes id="vm-guest1-meta_attributes">
                        <nvpair id="dummy2-remote-node"
                            name="remote-node" value="guest2"/>
                    </instance_attributes>
                </primitive>
                <primitive id="dummy3"
                        class="ocf" provider="pacemaker" type="remote">
                </primitive>
            </resources>
        """
        ).documentElement
        resources = dom.getElementsByTagName("resources")[0]
        resources.parentNode.replaceChild(new_resources, resources)

        self.assertEqual(
            None,
            utils.dom_get_resource_remote_node_name(
                utils.dom_get_resource(dom, "dummy1")
            ),
        )
        self.assertEqual(
            None,
            utils.dom_get_resource_remote_node_name(
                utils.dom_get_resource(dom, "dummy2")
            ),
        )
        self.assertEqual(
            "guest1",
            utils.dom_get_resource_remote_node_name(
                utils.dom_get_resource(dom, "vm-guest1")
            ),
        )
        self.assertEqual(
            "dummy3",
            utils.dom_get_resource_remote_node_name(
                utils.dom_get_resource(dom, "dummy3")
            ),
        )

    def test_dom_get_meta_attr_value(self):
        dom = self.get_cib_empty()
        new_resources = xml.dom.minidom.parseString(
            """
            <resources>
                <primitive id="dummy1"
                        class="ocf" provider="heartbeat" type="Dummy">
                </primitive>
                <primitive class="ocf" id="vm-guest1" provider="heartbeat"
                        type="VirtualDomain">
                    <instance_attributes id="vm-guest1-instance_attributes">
                        <nvpair id="vm-guest1-instance_attributes-hypervisor"
                            name="hypervisor" value="qemu:///system"/>
                        <nvpair id="vm-guest1-instance_attributes-config"
                            name="config" value="/root/guest1.xml"/>
                    </instance_attributes>
                    <meta_attributes id="vm-guest1-meta_attributes">
                        <nvpair id="vm-guest1-meta_attributes-remote-node"
                            name="remote-node" value="guest1"/>
                    </meta_attributes>
                </primitive>
                <primitive id="dummy2"
                        class="ocf" provider="heartbeat" type="Dummy">
                    <instance_attributes id="vm-guest1-meta_attributes">
                        <nvpair id="dummy2-remote-node"
                            name="remote-node" value="guest2"/>
                    </instance_attributes>
                </primitive>
            </resources>
        """
        ).documentElement
        resources = dom.getElementsByTagName("resources")[0]
        resources.parentNode.replaceChild(new_resources, resources)

        self.assertEqual(
            None,
            utils.dom_get_meta_attr_value(
                utils.dom_get_resource(dom, "dummy1"), "foo"
            ),
        )
        self.assertEqual(
            None,
            utils.dom_get_meta_attr_value(
                utils.dom_get_resource(dom, "dummy2"), "remote-node"
            ),
        )
        self.assertEqual(
            "guest1",
            utils.dom_get_meta_attr_value(
                utils.dom_get_resource(dom, "vm-guest1"), "remote-node"
            ),
        )
        self.assertEqual(
            None,
            utils.dom_get_meta_attr_value(
                utils.dom_get_resource(dom, "vm-guest1"), "foo"
            ),
        )

    def testGetElementWithId(self):
        dom = xml.dom.minidom.parseString(
            """
            <aa>
                <bb id="bb1"/>
                <bb/>
                <bb id="bb2">
                    <cc id="cc1"/>
                </bb>
                <bb id="bb3">
                    <cc id="cc2"/>
                </bb>
            </aa>
        """
        ).documentElement

        self.assert_element_id(
            utils.dom_get_element_with_id(dom, "bb", "bb1"), "bb1"
        )
        self.assert_element_id(
            utils.dom_get_element_with_id(dom, "bb", "bb2"), "bb2"
        )
        self.assert_element_id(
            utils.dom_get_element_with_id(dom, "cc", "cc1"), "cc1"
        )
        self.assert_element_id(
            utils.dom_get_element_with_id(
                utils.dom_get_element_with_id(dom, "bb", "bb2"), "cc", "cc1"
            ),
            "cc1",
        )
        self.assertEqual(None, utils.dom_get_element_with_id(dom, "dd", "bb1"))
        self.assertEqual(None, utils.dom_get_element_with_id(dom, "bb", "bb4"))
        self.assertEqual(None, utils.dom_get_element_with_id(dom, "bb", "cc1"))
        self.assertEqual(
            None,
            utils.dom_get_element_with_id(
                utils.dom_get_element_with_id(dom, "bb", "bb2"), "cc", "cc2"
            ),
        )

    def test_dom_get_parent_by_tag_name(self):
        dom = xml.dom.minidom.parseString(
            """
            <aa id="aa1">
                <bb id="bb1"/>
                <bb id="bb2">
                    <cc id="cc1"/>
                </bb>
                <bb id="bb3">
                    <cc id="cc2"/>
                </bb>
                <dd id="dd1" />
            </aa>
        """
        ).documentElement
        bb1 = utils.dom_get_element_with_id(dom, "bb", "bb1")
        cc1 = utils.dom_get_element_with_id(dom, "cc", "cc1")

        self.assert_element_id(
            utils.dom_get_parent_by_tag_names(bb1, ["aa"]), "aa1"
        )
        self.assert_element_id(
            utils.dom_get_parent_by_tag_names(cc1, ["aa"]), "aa1"
        )
        self.assert_element_id(
            utils.dom_get_parent_by_tag_names(cc1, ["bb"]), "bb2"
        )

        self.assertEqual(None, utils.dom_get_parent_by_tag_names(bb1, ["cc"]))
        self.assertEqual(None, utils.dom_get_parent_by_tag_names(cc1, ["dd"]))
        self.assertEqual(None, utils.dom_get_parent_by_tag_names(cc1, ["ee"]))

    def testValidateConstraintResource(self):
        dom = self.get_cib_resources()
        self.assertEqual(
            (True, "", "myClone"),
            utils.validate_constraint_resource(dom, "myClone"),
        )
        self.assertEqual(
            (True, "", "myGroupClone"),
            utils.validate_constraint_resource(dom, "myGroupClone"),
        )
        self.assertEqual(
            (True, "", "myMaster"),
            utils.validate_constraint_resource(dom, "myMaster"),
        )
        self.assertEqual(
            (True, "", "myGroupMaster"),
            utils.validate_constraint_resource(dom, "myGroupMaster"),
        )
        self.assertEqual(
            (True, "", "myBundle"),
            utils.validate_constraint_resource(dom, "myBundle"),
        )
        self.assertEqual(
            (True, "", "myEmptyBundle"),
            utils.validate_constraint_resource(dom, "myEmptyBundle"),
        )
        self.assertEqual(
            (True, "", "myResource"),
            utils.validate_constraint_resource(dom, "myResource"),
        )
        self.assertEqual(
            (True, "", "myGroup"),
            utils.validate_constraint_resource(dom, "myGroup"),
        )
        self.assertEqual(
            (True, "", "myGroupedResource"),
            utils.validate_constraint_resource(dom, "myGroupedResource"),
        )

        self.assertEqual(
            (False, "Resource 'myNonexistent' does not exist", None),
            utils.validate_constraint_resource(dom, "myNonexistent"),
        )

        message = (
            "%s is a clone resource, you should use the clone id: "
            "%s when adding constraints. Use --force to override."
        )
        self.assertEqual(
            (False, message % ("myClonedResource", "myClone"), "myClone"),
            utils.validate_constraint_resource(dom, "myClonedResource"),
        )
        self.assertEqual(
            (
                False,
                message % ("myClonedGroup", "myGroupClone"),
                "myGroupClone",
            ),
            utils.validate_constraint_resource(dom, "myClonedGroup"),
        )
        self.assertEqual(
            (
                False,
                message % ("myClonedGroupedResource", "myGroupClone"),
                "myGroupClone",
            ),
            utils.validate_constraint_resource(dom, "myClonedGroupedResource"),
        )

        message = (
            "%s is a clone resource, you should use the clone id: "
            "%s when adding constraints. Use --force to override."
        )
        self.assertEqual(
            (False, message % ("myMasteredResource", "myMaster"), "myMaster"),
            utils.validate_constraint_resource(dom, "myMasteredResource"),
        )
        self.assertEqual(
            (
                False,
                message % ("myMasteredGroup", "myGroupMaster"),
                "myGroupMaster",
            ),
            utils.validate_constraint_resource(dom, "myMasteredGroup"),
        )
        self.assertEqual(
            (
                False,
                message % ("myMasteredGroupedResource", "myGroupMaster"),
                "myGroupMaster",
            ),
            utils.validate_constraint_resource(
                dom, "myMasteredGroupedResource"
            ),
        )

        message = (
            "%s is a bundle resource, you should use the bundle id: "
            "%s when adding constraints. Use --force to override."
        )
        self.assertEqual(
            (False, message % ("myBundledResource", "myBundle"), "myBundle"),
            utils.validate_constraint_resource(dom, "myBundledResource"),
        )

        utils.pcs_options["--force"] = True
        self.assertEqual(
            (True, "", "myClone"),
            utils.validate_constraint_resource(dom, "myClonedResource"),
        )
        self.assertEqual(
            (True, "", "myGroupClone"),
            utils.validate_constraint_resource(dom, "myClonedGroup"),
        )
        self.assertEqual(
            (True, "", "myGroupClone"),
            utils.validate_constraint_resource(dom, "myClonedGroupedResource"),
        )
        self.assertEqual(
            (True, "", "myMaster"),
            utils.validate_constraint_resource(dom, "myMasteredResource"),
        )
        self.assertEqual(
            (True, "", "myGroupMaster"),
            utils.validate_constraint_resource(dom, "myMasteredGroup"),
        )
        self.assertEqual(
            (True, "", "myGroupMaster"),
            utils.validate_constraint_resource(
                dom, "myMasteredGroupedResource"
            ),
        )
        self.assertEqual(
            (True, "", "myBundle"),
            utils.validate_constraint_resource(dom, "myBundledResource"),
        )

    def testValidateXmlId(self):
        self.assertEqual((True, ""), utils.validate_xml_id("dummy"))
        self.assertEqual((True, ""), utils.validate_xml_id("DUMMY"))
        self.assertEqual((True, ""), utils.validate_xml_id("dUmMy"))
        self.assertEqual((True, ""), utils.validate_xml_id("dummy0"))
        self.assertEqual((True, ""), utils.validate_xml_id("dum0my"))
        self.assertEqual((True, ""), utils.validate_xml_id("dummy-"))
        self.assertEqual((True, ""), utils.validate_xml_id("dum-my"))
        self.assertEqual((True, ""), utils.validate_xml_id("dummy."))
        self.assertEqual((True, ""), utils.validate_xml_id("dum.my"))
        self.assertEqual((True, ""), utils.validate_xml_id("_dummy"))
        self.assertEqual((True, ""), utils.validate_xml_id("dummy_"))
        self.assertEqual((True, ""), utils.validate_xml_id("dum_my"))

        self.assertEqual(
            (False, "test id cannot be empty"),
            utils.validate_xml_id("", "test id"),
        )

        msg = "invalid test id '%s', '%s' is not a valid first character for a test id"
        self.assertEqual(
            (False, msg % ("0", "0")), utils.validate_xml_id("0", "test id")
        )
        self.assertEqual(
            (False, msg % ("-", "-")), utils.validate_xml_id("-", "test id")
        )
        self.assertEqual(
            (False, msg % (".", ".")), utils.validate_xml_id(".", "test id")
        )
        self.assertEqual(
            (False, msg % (":", ":")), utils.validate_xml_id(":", "test id")
        )
        self.assertEqual(
            (False, msg % ("0dummy", "0")),
            utils.validate_xml_id("0dummy", "test id"),
        )
        self.assertEqual(
            (False, msg % ("-dummy", "-")),
            utils.validate_xml_id("-dummy", "test id"),
        )
        self.assertEqual(
            (False, msg % (".dummy", ".")),
            utils.validate_xml_id(".dummy", "test id"),
        )
        self.assertEqual(
            (False, msg % (":dummy", ":")),
            utils.validate_xml_id(":dummy", "test id"),
        )

        msg = (
            "invalid test id '%s', '%s' is not a valid character for a test id"
        )
        self.assertEqual(
            (False, msg % ("dum:my", ":")),
            utils.validate_xml_id("dum:my", "test id"),
        )
        self.assertEqual(
            (False, msg % ("dummy:", ":")),
            utils.validate_xml_id("dummy:", "test id"),
        )
        self.assertEqual(
            (False, msg % ("dum?my", "?")),
            utils.validate_xml_id("dum?my", "test id"),
        )
        self.assertEqual(
            (False, msg % ("dummy?", "?")),
            utils.validate_xml_id("dummy?", "test id"),
        )

    def testIsIso8601Date(self):
        self.assertTrue(utils.is_iso8601_date("2014-07-03"))
        self.assertTrue(utils.is_iso8601_date("2014-07-03T11:35:14"))
        self.assertTrue(utils.is_iso8601_date("20140703"))
        self.assertTrue(utils.is_iso8601_date("2014-W27-4"))
        self.assertTrue(utils.is_iso8601_date("2014-184"))

        self.assertFalse(utils.is_iso8601_date(""))
        self.assertFalse(utils.is_iso8601_date("foo"))
        self.assertFalse(utils.is_iso8601_date("2014-07-32"))
        self.assertFalse(utils.is_iso8601_date("2014-13-03"))
        self.assertFalse(utils.is_iso8601_date("2014-W27-8"))
        self.assertFalse(utils.is_iso8601_date("2014-367"))

    def test_is_score(self):
        self.assertTrue(utils.is_score("INFINITY"))
        self.assertTrue(utils.is_score("+INFINITY"))
        self.assertTrue(utils.is_score("-INFINITY"))
        self.assertTrue(utils.is_score("0"))
        self.assertTrue(utils.is_score("+0"))
        self.assertTrue(utils.is_score("-0"))
        self.assertTrue(utils.is_score("123"))
        self.assertTrue(utils.is_score("-123"))
        self.assertTrue(utils.is_score("+123"))

        self.assertFalse(utils.is_score(""))
        self.assertFalse(utils.is_score("abc"))
        self.assertFalse(utils.is_score("+abc"))
        self.assertFalse(utils.is_score("-abc"))
        self.assertFalse(utils.is_score("10a"))
        self.assertFalse(utils.is_score("+10a"))
        self.assertFalse(utils.is_score("-10a"))
        self.assertFalse(utils.is_score("a10"))
        self.assertFalse(utils.is_score("+a10"))
        self.assertFalse(utils.is_score("a-10"))
        self.assertFalse(utils.is_score("infinity"))
        self.assertFalse(utils.is_score("+infinity"))
        self.assertFalse(utils.is_score("-infinity"))
        self.assertFalse(utils.is_score("+InFiNiTy"))
        self.assertFalse(utils.is_score("INFINITY10"))
        self.assertFalse(utils.is_score("INFINITY+10"))
        self.assertFalse(utils.is_score("-INFINITY10"))
        self.assertFalse(utils.is_score("+INFINITY+10"))
        self.assertFalse(utils.is_score("10INFINITY"))
        self.assertFalse(utils.is_score("+10+INFINITY"))

    def get_cib_status_lrm(self):
        cib_dom = self.get_cib_empty()
        new_status = xml.dom.minidom.parseString(
            """
<status>
  <node_state id="1" uname="rh70-node1">
    <lrm id="1">
      <lrm_resources>
        <lrm_resource id="dummy" type="Dummy" class="ocf" provider="heartbeat">
          <lrm_rsc_op id="dummy_monitor_30000" operation="monitor" call-id="34"
            rc-code="1" on_node="Xrh70-node1X" exit-reason="test" />
          <lrm_rsc_op id="dummy_stop_0" operation="stop" call-id="32"
            rc-code="0" />
          <lrm_rsc_op id="dummy_start_0" operation="start" call-id="33"
            rc-code="0" />
        </lrm_resource>
      </lrm_resources>
    </lrm>
  </node_state>
  <node_state id="2" uname="rh70-node2">
    <lrm id="2">
      <lrm_resources>
        <lrm_resource id="dummy" type="Dummy" class="ocf" provider="heartbeat">
          <lrm_rsc_op id="dummy_monitor_0" operation="monitor" call-id="5"
            rc-code="1" />
        </lrm_resource>
        <lrm_resource id="dummy1" type="Dummy" class="ocf" provider="heartbeat">
          <lrm_rsc_op id="dummy1_monitor_0" operation="monitor" call-id="3"
            rc-code="0" />
        </lrm_resource>
      </lrm_resources>
    </lrm>
  </node_state>
</status>
        """
        ).documentElement
        status = cib_dom.getElementsByTagName("status")[0]
        status.parentNode.replaceChild(new_status, status)
        return cib_dom

    def test_resource_running_on(self):
        status = xml.dom.minidom.parseString(
            """
<crm_mon>
    <summary />
    <nodes />
    <resources>
        <resource id="myResource" role="Started">
            <node name="rh70-node1" />
        </resource>
        <clone id="myClone">
            <resource id="myClonedResource" role="Started">
                <node name="rh70-node1" />
            </resource>
            <resource id="myClonedResource" role="Started">
                <node name="rh70-node2" />
            </resource>
            <resource id="myClonedResource" role="Started">
                <node name="rh70-node3" />
            </resource>
        </clone>
        <clone id="myMaster">
            <resource id="myMasteredResource:1" role="Slave">
                <node name="rh70-node2" />
            </resource>
            <resource id="myMasteredResource" role="Slave">
                <node name="rh70-node3" />
            </resource>
            <resource id="myMasteredResource" role="Master">
                <node name="rh70-node1" />
            </resource>
        </clone>
        <group id="myGroup">
             <resource id="myGroupedResource" role="Started">
                 <node name="rh70-node2" />
             </resource>
        </group>
        <clone id="myGroupClone">
            <group id="myClonedGroup:0">
                 <resource id="myClonedGroupedResource" role="Started">
                     <node name="rh70-node1" />
                 </resource>
            </group>
            <group id="myClonedGroup:1">
                 <resource id="myClonedGroupedResource" role="Started">
                     <node name="rh70-node2" />
                 </resource>
            </group>
            <group id="myClonedGroup:2">
                 <resource id="myClonedGroupedResource" role="Started">
                     <node name="rh70-node3" />
                 </resource>
            </group>
            <group id="myClonedGroup:3">
                 <resource id="myClonedGroupedResource" role="Started">
                     <node name="rh70-node3" />
                 </resource>
            </group>
        </clone>
        <clone id="myGroupMaster">
            <group id="myMasteredGroup:0">
                 <resource id="myMasteredGroupedResource" role="Slave">
                     <node name="rh70-node1" />
                 </resource>
            </group>
            <group id="myMasteredGroup:1">
                 <resource id="myMasteredGroupedResource" role="Master">
                     <node name="rh70-node2" />
                 </resource>
            </group>
            <group id="myMasteredGroup:2">
                 <resource id="myMasteredGroupedResource" role="Slave">
                     <node name="rh70-node3" />
                 </resource>
            </group>
        </clone>
        <resource id="myStoppedResource" role="Stopped">
        </resource>
    </resources>
</crm_mon>
        """
        ).documentElement

        self.assertEqual(
            utils.resource_running_on("myResource", status),
            {
                "message": "Resource 'myResource' is running on node rh70-node1.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myClonedResource", status),
            {
                "message": "Resource 'myClonedResource' is running on nodes "
                "rh70-node1, rh70-node2, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myClone", status),
            {
                "message": "Resource 'myClone' is running on nodes "
                "rh70-node1, rh70-node2, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myMasteredResource", status),
            {
                "message": "Resource 'myMasteredResource' is master on node "
                "rh70-node1; slave on nodes rh70-node2, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myMaster", status),
            {
                "message": "Resource 'myMaster' is master on node "
                "rh70-node1; slave on nodes rh70-node2, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myGroupedResource", status),
            {
                "message": "Resource 'myGroupedResource' is running on node "
                "rh70-node2.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myGroup", status),
            {
                "message": "Resource 'myGroup' is running on node "
                "rh70-node2.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myClonedGroupedResource", status),
            {
                "message": "Resource 'myClonedGroupedResource' is running on nodes "
                "rh70-node1, rh70-node2, rh70-node3, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myClonedGroup", status),
            {
                "message": "Resource 'myClonedGroup' is running on nodes "
                "rh70-node1, rh70-node2, rh70-node3, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myGroupClone", status),
            {
                "message": "Resource 'myGroupClone' is running on nodes "
                "rh70-node1, rh70-node2, rh70-node3, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myMasteredGroupedResource", status),
            {
                "message": "Resource 'myMasteredGroupedResource' is master on node "
                "rh70-node2; slave on nodes rh70-node1, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myMasteredGroup", status),
            {
                "message": "Resource 'myMasteredGroup' is master on node "
                "rh70-node2; slave on nodes rh70-node1, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myGroupMaster", status),
            {
                "message": "Resource 'myGroupMaster' is master on node "
                "rh70-node2; slave on nodes rh70-node1, rh70-node3.",
                "is_running": True,
            },
        )
        self.assertEqual(
            utils.resource_running_on("notMyResource", status),
            {
                "message": "Resource 'notMyResource' is not running on any node",
                "is_running": False,
            },
        )
        self.assertEqual(
            utils.resource_running_on("myStoppedResource", status),
            {
                "message": "Resource 'myStoppedResource' is not running on any node",
                "is_running": False,
            },
        )

    def test_get_operations_from_transitions(self):
        transitions = xml.dom.minidom.parse(rc("transitions01.xml"))
        self.assertEqual(
            [
                {
                    "id": "dummy",
                    "long_id": "dummy",
                    "operation": "stop",
                    "on_node": "rh7-3",
                },
                {
                    "id": "dummy",
                    "long_id": "dummy",
                    "operation": "start",
                    "on_node": "rh7-2",
                },
                {
                    "id": "d0",
                    "long_id": "d0:1",
                    "operation": "stop",
                    "on_node": "rh7-1",
                },
                {
                    "id": "d0",
                    "long_id": "d0:1",
                    "operation": "start",
                    "on_node": "rh7-2",
                },
                {
                    "id": "state",
                    "long_id": "state:0",
                    "operation": "stop",
                    "on_node": "rh7-3",
                },
                {
                    "id": "state",
                    "long_id": "state:0",
                    "operation": "start",
                    "on_node": "rh7-2",
                },
            ],
            utils.get_operations_from_transitions(transitions),
        )

        transitions = xml.dom.minidom.parse(rc("transitions02.xml"))
        self.assertEqual(
            [
                {
                    "id": "RemoteNode",
                    "long_id": "RemoteNode",
                    "operation": "stop",
                    "on_node": "virt-143",
                },
                {
                    "id": "RemoteNode",
                    "long_id": "RemoteNode",
                    "operation": "migrate_to",
                    "on_node": "virt-143",
                },
                {
                    "id": "RemoteNode",
                    "long_id": "RemoteNode",
                    "operation": "migrate_from",
                    "on_node": "virt-142",
                },
                {
                    "id": "dummy8",
                    "long_id": "dummy8",
                    "operation": "stop",
                    "on_node": "virt-143",
                },
                {
                    "id": "dummy8",
                    "long_id": "dummy8",
                    "operation": "start",
                    "on_node": "virt-142",
                },
            ],
            utils.get_operations_from_transitions(transitions),
        )

    def test_get_resources_location_from_operations(self):
        cib_dom = self.get_cib_resources()

        operations = []
        self.assertEqual(
            {},
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "start",
                "on_node": "rh7-1",
            },
        ]
        self.assertEqual(
            {
                "myResource": {
                    "id": "myResource",
                    "id_for_constraint": "myResource",
                    "long_id": "myResource",
                    "start_on_node": "rh7-1",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "start",
                "on_node": "rh7-1",
            },
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "start",
                "on_node": "rh7-2",
            },
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "monitor",
                "on_node": "rh7-3",
            },
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "stop",
                "on_node": "rh7-3",
            },
        ]
        self.assertEqual(
            {
                "myResource": {
                    "id": "myResource",
                    "id_for_constraint": "myResource",
                    "long_id": "myResource",
                    "start_on_node": "rh7-2",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "start",
                "on_node": "rh7-1",
            },
            {
                "id": "myClonedResource",
                "long_id": "myClonedResource:0",
                "operation": "start",
                "on_node": "rh7-1",
            },
            {
                "id": "myClonedResource",
                "long_id": "myClonedResource:0",
                "operation": "start",
                "on_node": "rh7-2",
            },
            {
                "id": "myClonedResource",
                "long_id": "myClonedResource:1",
                "operation": "start",
                "on_node": "rh7-3",
            },
        ]
        self.assertEqual(
            {
                "myResource": {
                    "id": "myResource",
                    "id_for_constraint": "myResource",
                    "long_id": "myResource",
                    "start_on_node": "rh7-1",
                },
                "myClonedResource:0": {
                    "id": "myClonedResource",
                    "id_for_constraint": "myClone",
                    "long_id": "myClonedResource:0",
                    "start_on_node": "rh7-2",
                },
                "myClonedResource:1": {
                    "id": "myClonedResource",
                    "id_for_constraint": "myClone",
                    "long_id": "myClonedResource:1",
                    "start_on_node": "rh7-3",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myUniqueClonedResource:0",
                "long_id": "myUniqueClonedResource:0",
                "operation": "start",
                "on_node": "rh7-1",
            },
            {
                "id": "myUniqueClonedResource:1",
                "long_id": "myUniqueClonedResource:1",
                "operation": "monitor",
                "on_node": "rh7-2",
            },
            {
                "id": "myUniqueClonedResource:2",
                "long_id": "myUniqueClonedResource:2",
                "operation": "start",
                "on_node": "rh7-3",
            },
        ]
        self.assertEqual(
            {
                "myUniqueClonedResource:0": {
                    "id": "myUniqueClonedResource:0",
                    "id_for_constraint": "myUniqueClone",
                    "long_id": "myUniqueClonedResource:0",
                    "start_on_node": "rh7-1",
                },
                "myUniqueClonedResource:2": {
                    "id": "myUniqueClonedResource:2",
                    "id_for_constraint": "myUniqueClone",
                    "long_id": "myUniqueClonedResource:2",
                    "start_on_node": "rh7-3",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myMasteredGroupedResource",
                "long_id": "myMasteredGroupedResource:0",
                "operation": "start",
                "on_node": "rh7-1",
            },
            {
                "id": "myMasteredGroupedResource",
                "long_id": "myMasteredGroupedResource:1",
                "operation": "demote",
                "on_node": "rh7-2",
            },
            {
                "id": "myMasteredGroupedResource",
                "long_id": "myMasteredGroupedResource:1",
                "operation": "promote",
                "on_node": "rh7-3",
            },
        ]
        self.assertEqual(
            {
                "myMasteredGroupedResource:0": {
                    "id": "myMasteredGroupedResource",
                    "id_for_constraint": "myGroupMaster",
                    "long_id": "myMasteredGroupedResource:0",
                    "start_on_node": "rh7-1",
                },
                "myMasteredGroupedResource:1": {
                    "id": "myMasteredGroupedResource",
                    "id_for_constraint": "myGroupMaster",
                    "long_id": "myMasteredGroupedResource:1",
                    "promote_on_node": "rh7-3",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

        operations = [
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "stop",
                "on_node": "rh7-1",
            },
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "migrate_to",
                "on_node": "rh7-1",
            },
            {
                "id": "myResource",
                "long_id": "myResource",
                "operation": "migrate_from",
                "on_node": "rh7-2",
            },
        ]
        self.assertEqual(
            {
                "myResource": {
                    "id": "myResource",
                    "id_for_constraint": "myResource",
                    "long_id": "myResource",
                    "start_on_node": "rh7-2",
                },
            },
            utils.get_resources_location_from_operations(cib_dom, operations),
        )

    def test_is_int(self):
        self.assertTrue(utils.is_int("-999"))
        self.assertTrue(utils.is_int("-1"))
        self.assertTrue(utils.is_int("0"))
        self.assertTrue(utils.is_int("1"))
        self.assertTrue(utils.is_int("99999"))
        self.assertTrue(utils.is_int(" 99999  "))
        self.assertFalse(utils.is_int("0.0"))
        self.assertFalse(utils.is_int("-1.0"))
        self.assertFalse(utils.is_int("-0.1"))
        self.assertFalse(utils.is_int("0.001"))
        self.assertFalse(utils.is_int("-999999.1"))
        self.assertFalse(utils.is_int("0.0001"))
        self.assertFalse(utils.is_int(""))
        self.assertFalse(utils.is_int("   "))
        self.assertFalse(utils.is_int("A"))
        self.assertFalse(utils.is_int("random 15 47 text  "))

    def test_dom_get_node(self):
        cib = self.get_cib_with_nodes_minidom()
        self.assertIsNone(utils.dom_get_node(cib, "non-existing-node"))
        node = utils.dom_get_node(cib, "rh7-1")
        self.assertEqual(node.getAttribute("uname"), "rh7-1")
        self.assertEqual(node.getAttribute("id"), "1")

    def test_dom_prepare_child_element(self):
        cib = self.get_cib_with_nodes_minidom()
        node = cib.getElementsByTagName("node")[0]
        self.assertEqual(len(dom_get_child_elements(node)), 0)
        child = utils.dom_prepare_child_element(
            node, "utilization", "rh7-1-utilization"
        )
        self.assertEqual(len(dom_get_child_elements(node)), 1)
        self.assertEqual(child, dom_get_child_elements(node)[0])
        self.assertEqual(dom_get_child_elements(node)[0].tagName, "utilization")
        self.assertEqual(
            dom_get_child_elements(node)[0].getAttribute("id"),
            "rh7-1-utilization",
        )
        child2 = utils.dom_prepare_child_element(
            node, "utilization", "rh7-1-utilization"
        )
        self.assertEqual(len(dom_get_child_elements(node)), 1)
        self.assertEqual(child, child2)

    def test_dom_update_nv_pair_add(self):
        nv_set = xml.dom.minidom.parseString("<nvset/>").documentElement
        utils.dom_update_nv_pair(nv_set, "test_name", "test_val", "prefix-")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 1)
        pair = dom_get_child_elements(nv_set)[0]
        self.assertEqual(pair.getAttribute("name"), "test_name")
        self.assertEqual(pair.getAttribute("value"), "test_val")
        self.assertEqual(pair.getAttribute("id"), "prefix-test_name")
        utils.dom_update_nv_pair(nv_set, "another_name", "value", "prefix2-")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 2)
        self.assertEqual(pair, dom_get_child_elements(nv_set)[0])
        pair = dom_get_child_elements(nv_set)[1]
        self.assertEqual(pair.getAttribute("name"), "another_name")
        self.assertEqual(pair.getAttribute("value"), "value")
        self.assertEqual(pair.getAttribute("id"), "prefix2-another_name")

    def test_dom_update_nv_pair_update(self):
        nv_set = xml.dom.minidom.parseString(
            """
        <nv_set>
            <nvpair id="prefix-test_name" name="test_name" value="test_val"/>
            <nvpair id="prefix2-another_name" name="another_name" value="value"/>
        </nv_set>
        """
        ).documentElement
        utils.dom_update_nv_pair(nv_set, "test_name", "new_value")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 2)
        pair1 = dom_get_child_elements(nv_set)[0]
        pair2 = dom_get_child_elements(nv_set)[1]
        self.assertEqual(pair1.getAttribute("name"), "test_name")
        self.assertEqual(pair1.getAttribute("value"), "new_value")
        self.assertEqual(pair1.getAttribute("id"), "prefix-test_name")
        self.assertEqual(pair2.getAttribute("name"), "another_name")
        self.assertEqual(pair2.getAttribute("value"), "value")
        self.assertEqual(pair2.getAttribute("id"), "prefix2-another_name")

    def test_dom_update_nv_pair_remove(self):
        nv_set = xml.dom.minidom.parseString(
            """
        <nv_set>
            <nvpair id="prefix-test_name" name="test_name" value="test_val"/>
            <nvpair id="prefix2-another_name" name="another_name" value="value"/>
        </nv_set>
        """
        ).documentElement
        utils.dom_update_nv_pair(nv_set, "non_existing_name", "")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 2)
        utils.dom_update_nv_pair(nv_set, "another_name", "")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 1)
        pair = dom_get_child_elements(nv_set)[0]
        self.assertEqual(pair.getAttribute("name"), "test_name")
        self.assertEqual(pair.getAttribute("value"), "test_val")
        self.assertEqual(pair.getAttribute("id"), "prefix-test_name")
        utils.dom_update_nv_pair(nv_set, "test_name", "")
        self.assertEqual(len(dom_get_child_elements(nv_set)), 0)

    def test_convert_args_to_tuples(self):
        out = utils.convert_args_to_tuples(
            ["invalid_string", "key=value", "key2=val=ue", "k e y= v a l u e "]
        )
        self.assertEqual(
            out,
            [("key", "value"), ("key2", "val=ue"), ("k e y", " v a l u e ")],
        )

    def test_dom_update_utilization_invalid(self):
        # commands writes to stderr
        # we want clean test output, so we capture it
        tmp_stderr = sys.stderr
        sys.stderr = StringIO()

        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id"/>
        """
        ).documentElement
        self.assertRaises(
            SystemExit,
            utils.dom_update_utilization,
            el,
            {"name": "invalid_val"},
        )

        self.assertRaises(
            SystemExit, utils.dom_update_utilization, el, {"name": "0.01"}
        )

        sys.stderr = tmp_stderr

    def test_dom_update_utilization_add(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id"/>
        """
        ).documentElement
        utils.dom_update_utilization(
            el,
            {
                "name": "",
                "key": "-1",
                "keys": "90",
            },
        )

        self.assertEqual(len(dom_get_child_elements(el)), 1)
        u = dom_get_child_elements(el)[0]
        self.assertEqual(u.tagName, "utilization")
        self.assertEqual(u.getAttribute("id"), "test_id-utilization")
        self.assertEqual(len(dom_get_child_elements(u)), 2)

        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("id"),
            "test_id-utilization-key",
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("name"), "key"
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("value"), "-1"
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("id"),
            "test_id-utilization-keys",
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("name"), "keys"
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("value"), "90"
        )

    def test_dom_update_utilization_update_remove(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id">
            <utilization id="test_id-utilization">
                <nvpair id="test_id-utilization-key" name="key" value="-1"/>
                <nvpair id="test_id-utilization-keys" name="keys" value="90"/>
            </utilization>
        </resource>
        """
        ).documentElement
        utils.dom_update_utilization(
            el,
            {
                "key": "100",
                "keys": "",
            },
        )

        u = dom_get_child_elements(el)[0]
        self.assertEqual(len(dom_get_child_elements(u)), 1)
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("id"),
            "test_id-utilization-key",
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("name"), "key"
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("value"), "100"
        )

    def test_dom_update_meta_attr_add(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id"/>
        """
        ).documentElement
        utils.dom_update_meta_attr(
            el, [("name", ""), ("key", "test"), ("key2", "val")]
        )

        self.assertEqual(len(dom_get_child_elements(el)), 1)
        u = dom_get_child_elements(el)[0]
        self.assertEqual(u.tagName, "meta_attributes")
        self.assertEqual(u.getAttribute("id"), "test_id-meta_attributes")
        self.assertEqual(len(dom_get_child_elements(u)), 2)

        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("id"),
            "test_id-meta_attributes-key",
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("name"), "key"
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("value"), "test"
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("id"),
            "test_id-meta_attributes-key2",
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("name"), "key2"
        )
        self.assertEqual(
            dom_get_child_elements(u)[1].getAttribute("value"), "val"
        )

    def test_dom_update_meta_attr_update_remove(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id">
            <meta_attributes id="test_id-utilization">
                <nvpair id="test_id-meta_attributes-key" name="key" value="test"/>
                <nvpair id="test_id-meta_attributes-key2" name="key2" value="val"/>
            </meta_attributes>
        </resource>
        """
        ).documentElement
        utils.dom_update_meta_attr(el, [("key", "another_val"), ("key2", "")])

        u = dom_get_child_elements(el)[0]
        self.assertEqual(len(dom_get_child_elements(u)), 1)
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("id"),
            "test_id-meta_attributes-key",
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("name"), "key"
        )
        self.assertEqual(
            dom_get_child_elements(u)[0].getAttribute("value"), "another_val"
        )

    def test_get_utilization(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id">
            <utilization id="test_id-utilization">
                <nvpair id="test_id-utilization-key" name="key" value="-1"/>
                <nvpair id="test_id-utilization-keys" name="keys" value="90"/>
            </utilization>
        </resource>
        """
        ).documentElement
        self.assertEqual({"key": "-1", "keys": "90"}, utils.get_utilization(el))

    def test_get_utilization_str(self):
        el = xml.dom.minidom.parseString(
            """
        <resource id="test_id">
            <utilization id="test_id-utilization">
                <nvpair id="test_id-utilization-key" name="key" value="-1"/>
                <nvpair id="test_id-utilization-keys" name="keys" value="90"/>
            </utilization>
        </resource>
        """
        ).documentElement
        self.assertEqual("key=-1 keys=90", utils.get_utilization_str(el))

    def test_get_cluster_property_from_xml_enum(self):
        el = ET.fromstring(
            """
        <parameter name="no-quorum-policy" unique="0">
            <shortdesc lang="en">What to do when the cluster does not have quorum</shortdesc>
            <content type="enum" default="stop"/>
            <longdesc lang="en">What to do when the cluster does not have quorum  Allowed values: stop, freeze, ignore, suicide</longdesc>
        </parameter>
        """
        )
        expected = {
            "name": "no-quorum-policy",
            "shortdesc": "What to do when the cluster does not have quorum",
            "longdesc": "",
            "type": "enum",
            "default": "stop",
            "enum": ["stop", "freeze", "ignore", "suicide"],
        }
        self.assertEqual(expected, utils.get_cluster_property_from_xml(el))

    def test_get_cluster_property_from_xml(self):
        el = ET.fromstring(
            """
        <parameter name="an-integer-property" unique="0">
            <shortdesc lang="en"></shortdesc>
            <content type="integer" default="0"/>
            <longdesc lang="en"></longdesc>
        </parameter>
        """
        )
        expected = {
            "name": "an-integer-property",
            "shortdesc": "",
            "longdesc": "",
            "type": "integer",
            "default": "0",
        }
        self.assertEqual(expected, utils.get_cluster_property_from_xml(el))

    def test_is_valid_cib_value_unknown_type(self):
        # should be always true
        self.assertTrue(utils.is_valid_cib_value("unknown", "test"))
        self.assertTrue(utils.is_valid_cib_value("string", "string value"))

    def test_is_valid_cib_value_integer(self):
        self.assertTrue(utils.is_valid_cib_value("integer", "0"))
        self.assertTrue(utils.is_valid_cib_value("integer", "42"))
        self.assertTrue(utils.is_valid_cib_value("integer", "-90"))
        self.assertTrue(utils.is_valid_cib_value("integer", "+90"))
        self.assertTrue(utils.is_valid_cib_value("integer", "INFINITY"))
        self.assertTrue(utils.is_valid_cib_value("integer", "-INFINITY"))
        self.assertTrue(utils.is_valid_cib_value("integer", "+INFINITY"))
        self.assertFalse(utils.is_valid_cib_value("integer", "0.0"))
        self.assertFalse(utils.is_valid_cib_value("integer", "-10.9"))
        self.assertFalse(utils.is_valid_cib_value("integer", "string"))

    def test_is_valid_cib_value_enum(self):
        self.assertTrue(
            utils.is_valid_cib_value("enum", "this", ["another", "this", "1"])
        )
        self.assertFalse(
            utils.is_valid_cib_value("enum", "this", ["another", "this_not"])
        )
        self.assertFalse(utils.is_valid_cib_value("enum", "this", []))
        self.assertFalse(utils.is_valid_cib_value("enum", "this"))

    def test_is_valid_cib_value_boolean(self):
        self.assertTrue(utils.is_valid_cib_value("boolean", "true"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "TrUe"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "TRUE"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "yes"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "on"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "y"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "Y"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "1"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "false"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "FaLse"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "FALSE"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "off"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "no"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "N"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "n"))
        self.assertTrue(utils.is_valid_cib_value("boolean", "0"))
        self.assertFalse(utils.is_valid_cib_value("boolean", "-1"))
        self.assertFalse(utils.is_valid_cib_value("boolean", "not"))
        self.assertFalse(utils.is_valid_cib_value("boolean", "random_string"))
        self.assertFalse(utils.is_valid_cib_value("boolean", "truth"))

    def test_is_valid_cib_value_time(self):
        self.assertTrue(utils.is_valid_cib_value("time", "10"))
        self.assertTrue(utils.is_valid_cib_value("time", "0"))
        self.assertTrue(utils.is_valid_cib_value("time", "9s"))
        self.assertTrue(utils.is_valid_cib_value("time", "10sec"))
        self.assertTrue(utils.is_valid_cib_value("time", "10min"))
        self.assertTrue(utils.is_valid_cib_value("time", "10m"))
        self.assertTrue(utils.is_valid_cib_value("time", "10h"))
        self.assertTrue(utils.is_valid_cib_value("time", "10hr"))
        self.assertFalse(utils.is_valid_cib_value("time", "5.2"))
        self.assertFalse(utils.is_valid_cib_value("time", "-10"))
        self.assertFalse(utils.is_valid_cib_value("time", "10m 2s"))
        self.assertFalse(utils.is_valid_cib_value("time", "hour"))
        self.assertFalse(utils.is_valid_cib_value("time", "day"))

    def test_validate_cluster_property(self):
        definition = {
            "property-int": {
                "name": "property-int",
                "shortdesc": "int description",
                "longdesc": "long int description",
                "type": "integer",
                "default": "0",
                "source": "pacemaker-schedulerd",
            },
            "property-enum": {
                "name": "property-enum",
                "shortdesc": "What to do when the cluster does not have quorum",
                "longdesc": (
                    "What to do when the cluster does not have quorum  "
                    "Allowed values: stop, freeze, ignore, suicide"
                ),
                "type": "enum",
                "default": "stop",
                "enum": ["stop", "freeze", "ignore", "suicide"],
                "source": "pacemaker-schedulerd",
            },
            "property-bool": {
                "name": "property-bool",
                "shortdesc": "Enable CIB ACL",
                "longdesc": "Enable CIB ACL",
                "type": "boolean",
                "default": "false",
                "source": "cib",
            },
        }
        self.assertTrue(
            utils.is_valid_cluster_property(definition, "property-int", "10")
        )
        self.assertTrue(
            utils.is_valid_cluster_property(definition, "property-int", "-1")
        )
        self.assertTrue(
            utils.is_valid_cluster_property(
                definition, "property-enum", "freeze"
            )
        )
        self.assertTrue(
            utils.is_valid_cluster_property(
                definition, "property-enum", "suicide"
            )
        )
        self.assertTrue(
            utils.is_valid_cluster_property(definition, "property-bool", "true")
        )
        self.assertTrue(
            utils.is_valid_cluster_property(
                definition, "property-bool", "false"
            )
        )
        self.assertTrue(
            utils.is_valid_cluster_property(definition, "property-bool", "on")
        )
        self.assertTrue(
            utils.is_valid_cluster_property(definition, "property-bool", "OFF")
        )
        self.assertFalse(
            utils.is_valid_cluster_property(definition, "property-int", "test")
        )
        self.assertFalse(
            utils.is_valid_cluster_property(definition, "property-int", "1.2")
        )
        self.assertFalse(
            utils.is_valid_cluster_property(
                definition, "property-enum", "invalid"
            )
        )
        self.assertFalse(
            utils.is_valid_cluster_property(definition, "property-bool", "not")
        )
        self.assertRaises(
            utils.UnknownPropertyException,
            utils.is_valid_cluster_property,
            definition,
            "unknown",
            "value",
        )

    def assert_element_id(self, node, node_id, tag=None):
        self.assertTrue(
            isinstance(node, xml.dom.minidom.Element),
            "element with id '%s' not found" % node_id,
        )
        self.assertEqual(node.getAttribute("id"), node_id)
        if tag:
            self.assertEqual(node.tagName, tag)


class RunParallelTest(TestCase):
    @staticmethod
    def fixture_create_worker(log, name, sleepSeconds=0):
        def worker():
            sleep(sleepSeconds)
            log.append(name)

        return worker

    def test_run_all_workers(self):
        log = []
        utils.run_parallel(
            [
                self.fixture_create_worker(log, "first"),
                self.fixture_create_worker(log, "second"),
            ],
            wait_seconds=0.1,
        )

        self.assertEqual(sorted(log), sorted(["first", "second"]))


class NodeActionTaskTest(TestCase):
    def test_can_run_action(self):
        def action(node, arg, kwarg=None):
            return (0, ":".join([node, arg, kwarg]))

        report_list = []

        def report(node, returncode, output):
            report_list.append("|".join([node, str(returncode), output]))

        task = utils.create_task(report, action, "node", "arg", kwarg="kwarg")
        task()

        self.assertEqual(["node|0|node:arg:kwarg"], report_list)


class TouchCibFile(TestCase):
    @mock.patch("pcs.utils.os.path.isfile", mock.Mock(return_value=False))
    @mock.patch(
        "pcs.utils.write_empty_cib",
        mock.Mock(side_effect=EnvironmentError("some message")),
    )
    @mock.patch("pcs.utils.err")
    def test_exception_is_transformed_correctly(self, err):
        # pylint: disable=no-self-use
        filename = "/fake/filename"
        utils.touch_cib_file(filename)
        err.assert_called_once_with(
            "Unable to write to file: '/fake/filename': 'some message'"
        )



from functools import partial
import logging
from unittest import mock, TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import create_patcher
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common.reports import codes as report_codes
from pcs.common.fencing_topology import (
    TARGET_TYPE_ATTRIBUTE,
    TARGET_TYPE_NODE,
    TARGET_TYPE_REGEXP,
)
from pcs.lib.env import LibraryEnvironment

from pcs.lib.commands import fencing_topology as lib

# pylint: disable=no-self-use

create_lib_env = partial(
    LibraryEnvironment,
    mock.MagicMock(logging.Logger),
    MockLibraryReportProcessor(),
)
patch_env = partial(mock.patch.object, LibraryEnvironment)
patch_command = create_patcher("pcs.lib.commands.fencing_topology")


@patch_command("cib_fencing_topology.add_level")
@patch_command("get_resources")
@patch_command("get_fencing_topology")
@patch_env("push_cib")
@patch_command("ClusterState")
@patch_env("get_cluster_state")
@patch_env("get_cib")
class AddLevel(TestCase):
    def prepare_mocks(
        self,
        mock_get_cib,
        mock_status_dom,
        mock_status,
        mock_get_topology,
        mock_get_resources,
    ):
        mock_get_cib.return_value = "mocked cib"
        mock_status_dom.return_value = "mock get_cluster_status_dom"
        mock_status.return_value = mock.MagicMock(
            node_section=mock.MagicMock(nodes="nodes")
        )
        mock_get_topology.return_value = "topology el"
        mock_get_resources.return_value = "resources_el"

    def assert_mocks(
        self,
        mock_status_dom,
        mock_status,
        mock_get_topology,
        mock_get_resources,
        mock_push_cib,
    ):
        mock_status_dom.assert_called_once_with()
        mock_status.assert_called_once_with("mock get_cluster_status_dom")
        mock_get_topology.assert_called_once_with("mocked cib")
        mock_get_resources.assert_called_once_with("mocked cib")
        mock_push_cib.assert_called_once_with()

    def test_success(
        self,
        mock_get_cib,
        mock_status_dom,
        mock_status,
        mock_push_cib,
        mock_get_topology,
        mock_get_resources,
        mock_add_level,
    ):
        self.prepare_mocks(
            mock_get_cib,
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
        )
        lib_env = create_lib_env()

        lib.add_level(
            lib_env,
            "level",
            "target type",
            "target value",
            "devices",
            "force device",
            "force node",
        )

        mock_add_level.assert_called_once_with(
            lib_env.report_processor,
            "topology el",
            "resources_el",
            "level",
            "target type",
            "target value",
            "devices",
            "nodes",
            "force device",
            "force node",
        )
        mock_get_cib.assert_called_once_with()
        self.assert_mocks(
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
            mock_push_cib,
        )

    def test_target_attribute_updates_cib(
        self,
        mock_get_cib,
        mock_status_dom,
        mock_status,
        mock_push_cib,
        mock_get_topology,
        mock_get_resources,
        mock_add_level,
    ):
        self.prepare_mocks(
            mock_get_cib,
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
        )
        lib_env = create_lib_env()

        lib.add_level(
            lib_env,
            "level",
            TARGET_TYPE_ATTRIBUTE,
            "target value",
            "devices",
            "force device",
            "force node",
        )

        mock_add_level.assert_called_once_with(
            lib_env.report_processor,
            "topology el",
            "resources_el",
            "level",
            TARGET_TYPE_ATTRIBUTE,
            "target value",
            "devices",
            "nodes",
            "force device",
            "force node",
        )
        mock_get_cib.assert_called_once_with()
        self.assert_mocks(
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
            mock_push_cib,
        )

    def test_target_regexp_updates_cib(
        self,
        mock_get_cib,
        mock_status_dom,
        mock_status,
        mock_push_cib,
        mock_get_topology,
        mock_get_resources,
        mock_add_level,
    ):
        self.prepare_mocks(
            mock_get_cib,
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
        )
        lib_env = create_lib_env()

        lib.add_level(
            lib_env,
            "level",
            TARGET_TYPE_REGEXP,
            "target value",
            "devices",
            "force device",
            "force node",
        )

        mock_add_level.assert_called_once_with(
            lib_env.report_processor,
            "topology el",
            "resources_el",
            "level",
            TARGET_TYPE_REGEXP,
            "target value",
            "devices",
            "nodes",
            "force device",
            "force node",
        )
        mock_get_cib.assert_called_once_with()
        self.assert_mocks(
            mock_status_dom,
            mock_status,
            mock_get_topology,
            mock_get_resources,
            mock_push_cib,
        )


@patch_command("cib_fencing_topology.export")
@patch_command("get_fencing_topology")
@patch_env("push_cib")
@patch_env("get_cib", lambda self: "mocked cib")
class GetConfig(TestCase):
    def test_success(self, mock_push_cib, mock_get_topology, mock_export):
        mock_get_topology.return_value = "topology el"
        mock_export.return_value = "exported config"
        lib_env = create_lib_env()

        self.assertEqual("exported config", lib.get_config(lib_env))

        mock_export.assert_called_once_with("topology el")
        mock_get_topology.assert_called_once_with("mocked cib")
        mock_push_cib.assert_not_called()


@patch_command("cib_fencing_topology.remove_all_levels")
@patch_command("get_fencing_topology")
@patch_env("push_cib")
@patch_env("get_cib", lambda self: "mocked cib")
class RemoveAllLevels(TestCase):
    def test_success(self, mock_push_cib, mock_get_topology, mock_remove):
        mock_get_topology.return_value = "topology el"
        lib_env = create_lib_env()

        lib.remove_all_levels(lib_env)

        mock_remove.assert_called_once_with("topology el")
        mock_get_topology.assert_called_once_with("mocked cib")
        mock_push_cib.assert_called_once_with()


class RemoveLevelsByParams(TestCase):
    fixture_cib_node = """
        <fencing-topology>
            <fencing-level
                id="fl1" index="1" devices="dev1,dev2" target="node1"
            />
            <fencing-level
                id="flX" index="2" devices="devX,devY" target="node1"
            />
        </fencing-topology>
    """

    fixture_cib_attr = """
        <fencing-topology>
            <fencing-level
                id="fl1" index="1" devices="dev1,dev2"
                target-attribute="fencing" target-value="improved"
            />
            <fencing-level
                id="flX" index="2" devices="devX,devY" target="node1"
            />
        </fencing-topology>
    """

    fixture_cib_regexp = """
        <fencing-topology>
            <fencing-level
                id="fl1" index="1" devices="dev1,dev2" target-pattern="node\\d+"
            />
            <fencing-level
                id="flX" index="2" devices="devX,devY" target="node1"
            />
        </fencing-topology>
    """

    fixture_cib_other_level = """
        <fencing-topology>
            <fencing-level
                id="flX" index="2" devices="devX,devY" target="node1"
            />
        </fencing-topology>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_devices_target_attr_present(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_attr)
        self.config.env.push_cib(fencing_topology=self.fixture_cib_other_level)

        lib.remove_levels_by_params(
            self.env_assist.get_env(),
            level="1",
            target_type=TARGET_TYPE_ATTRIBUTE,
            target_value=("fencing", "improved"),
            devices=["dev1", "dev2"],
        )

    def test_devices_target_attr_missing(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_attr)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.env_assist.get_env(),
                level="2",
                target_type=TARGET_TYPE_ATTRIBUTE,
                target_value=("fencing", "improved"),
                devices=["dev1", "dev2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    level="2",
                    target_type=TARGET_TYPE_ATTRIBUTE,
                    target_value=("fencing", "improved"),
                    devices=["dev1", "dev2"],
                ),
            ]
        )

    def test_devices_target_regexp_present(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_regexp)
        self.config.env.push_cib(fencing_topology=self.fixture_cib_other_level)

        lib.remove_levels_by_params(
            self.env_assist.get_env(),
            level="1",
            target_type=TARGET_TYPE_REGEXP,
            target_value="node\\d+",
            devices=["dev1", "dev2"],
        )

    def test_devices_target_regexp_missing(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_regexp)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.env_assist.get_env(),
                level="2",
                target_type=TARGET_TYPE_REGEXP,
                target_value="node\\d+",
                devices=["dev1", "dev2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    level="2",
                    target_type=TARGET_TYPE_REGEXP,
                    target_value="node\\d+",
                    devices=["dev1", "dev2"],
                ),
            ]
        )

    def test_devices_target_node_present(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_node)
        self.config.env.push_cib(fencing_topology=self.fixture_cib_other_level)

        lib.remove_levels_by_params(
            self.env_assist.get_env(),
            level="1",
            target_type=TARGET_TYPE_NODE,
            target_value="node1",
            devices=["dev1", "dev2"],
        )

    def test_devices_target_node_missing_success(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_node)
        self.config.env.push_cib(fencing_topology=self.fixture_cib_other_level)

        lib.remove_levels_by_params(
            self.env_assist.get_env(),
            level="1",
            target_type=TARGET_TYPE_NODE,
            target_value="dev1",
            devices=["dev2"],
            target_may_be_a_device=True,
        )

    def test_devices_target_node_missing_not_found(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_node)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.env_assist.get_env(),
                level="2",
                target_type=TARGET_TYPE_NODE,
                target_value="dev1",
                devices=["dev2"],
                target_may_be_a_device=True,
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    level="2",
                    target_type=TARGET_TYPE_NODE,
                    target_value="dev1",
                    devices=["dev2"],
                ),
                fixture.error(
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    level="2",
                    target_type=None,
                    target_value=None,
                    devices=["dev1", "dev2"],
                ),
            ]
        )

    def test_devices_target_node_missing_guessing_disabled(self):
        self.config.runner.cib.load(fencing_topology=self.fixture_cib_node)

        self.env_assist.assert_raise_library_error(
            lambda: lib.remove_levels_by_params(
                self.env_assist.get_env(),
                level="1",
                target_type=TARGET_TYPE_NODE,
                target_value="dev1",
                devices=["dev2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST,
                    level="1",
                    target_type=TARGET_TYPE_NODE,
                    target_value="dev1",
                    devices=["dev2"],
                ),
            ]
        )


@patch_command("cib_fencing_topology.verify")
@patch_command("get_resources")
@patch_command("get_fencing_topology")
@patch_env("push_cib")
@patch_command("ClusterState")
@patch_env("get_cluster_state")
@patch_env("get_cib", lambda self: "mocked cib")
class Verify(TestCase):
    def test_success(
        self,
        mock_status_dom,
        mock_status,
        mock_push_cib,
        mock_get_topology,
        mock_get_resources,
        mock_verify,
    ):
        mock_status_dom.return_value = "mock get_cluster_status_dom"
        mock_status.return_value = mock.MagicMock(
            node_section=mock.MagicMock(nodes="nodes")
        )
        mock_get_topology.return_value = "topology el"
        mock_get_resources.return_value = "resources_el"
        lib_env = create_lib_env()

        lib.verify(lib_env)

        mock_verify.assert_called_once_with(
            "topology el", "resources_el", "nodes"
        )
        mock_status_dom.assert_called_once_with()
        mock_status.assert_called_once_with("mock get_cluster_status_dom")
        mock_get_topology.assert_called_once_with("mocked cib")
        mock_get_resources.assert_called_once_with("mocked cib")
        mock_push_cib.assert_not_called()



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs.common.reports import codes as report_codes
from pcs.lib.commands import resource


class GetFailcounts(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    @staticmethod
    def fixture_cib():
        return """
        <cib>
            <status>
                <node_state uname="node1">
                    <transient_attributes>
                        <instance_attributes>
                            <nvpair name="fail-count-resource#start_0"
                                value="INFINITY"/>
                            <nvpair name="last-failure-resource#start_0"
                                value="1528871936"/>
                        </instance_attributes>
                    </transient_attributes>
                </node_state>
                <node_state uname="node2">
                    <transient_attributes>
                        <instance_attributes>
                            <nvpair name="fail-count-resource#monitor_5000"
                                value="10"/>
                            <nvpair name="last-failure-resource#monitor_5000"
                                value="1528871946"/>
                        </instance_attributes>
                    </transient_attributes>
                </node_state>
            </status>
        </cib>
        """

    def test_operation_requires_interval(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.get_failcounts(
                self.env_assist.get_env(), interval="1000"
            ),
            [
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name="interval",
                    option_type=None,
                    prerequisite_name="operation",
                    prerequisite_type=None,
                ),
            ],
            expected_in_processor=False,
        )

    def test_bad_interval(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.get_failcounts(
                self.env_assist.get_env(), operation="start", interval="often"
            ),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="interval",
                    option_value="often",
                    allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
            expected_in_processor=False,
        )

    def test_all_validation_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.get_failcounts(
                self.env_assist.get_env(), interval="often"
            ),
            [
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_IS_MISSING,
                    option_name="interval",
                    option_type=None,
                    prerequisite_name="operation",
                    prerequisite_type=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="interval",
                    option_value="often",
                    allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
            expected_in_processor=False,
        )

    def test_get_all(self):
        self.config.runner.cib.load_content(self.fixture_cib())
        self.assertEqual(
            resource.get_failcounts(self.env_assist.get_env()),
            [
                {
                    "node": "node1",
                    "resource": "resource",
                    "clone_id": None,
                    "operation": "start",
                    "interval": "0",
                    "fail_count": "INFINITY",
                    "last_failure": 1528871936,
                },
                {
                    "node": "node2",
                    "resource": "resource",
                    "clone_id": None,
                    "operation": "monitor",
                    "interval": "5000",
                    "fail_count": 10,
                    "last_failure": 1528871946,
                },
            ],
        )

    def test_filter_node(self):
        self.config.runner.cib.load_content(self.fixture_cib())
        self.assertEqual(
            resource.get_failcounts(self.env_assist.get_env(), node="node2"),
            [
                {
                    "node": "node2",
                    "resource": "resource",
                    "clone_id": None,
                    "operation": "monitor",
                    "interval": "5000",
                    "fail_count": 10,
                    "last_failure": 1528871946,
                },
            ],
        )

    def test_filter_interval(self):
        self.config.runner.cib.load_content(self.fixture_cib())
        self.assertEqual(
            resource.get_failcounts(
                self.env_assist.get_env(), operation="monitor", interval="5"
            ),
            [
                {
                    "node": "node2",
                    "resource": "resource",
                    "clone_id": None,
                    "operation": "monitor",
                    "interval": "5000",
                    "fail_count": 10,
                    "last_failure": 1528871946,
                },
            ],
        )



from pcs_test.tools.assertions import prepare_diff

from pcs.lib.corosync.config_facade import ConfigFacade
from pcs.lib.errors import LibraryError

CALL_TYPE_PUSH_COROSYNC_CONF = "CALL_TYPE_PUSH_COROSYNC_CONF"


class Call:
    type = CALL_TYPE_PUSH_COROSYNC_CONF

    def __init__(
        self,
        corosync_conf_text,
        skip_offline_targets,
        raises=False,
        need_stopped_cluster=False,
    ):
        self.corosync_conf_text = corosync_conf_text
        self.skip_offline_targets = skip_offline_targets
        self.raises = raises
        self.need_stopped_cluster = need_stopped_cluster

    def __repr__(self):
        return str("<CorosyncConfPush skip-offline='{0}'>").format(
            self.skip_offline_targets
        )


def get_push_corosync_conf(call_queue):
    def push_corosync_conf(
        lib_env, corosync_conf_facade, skip_offline_nodes=False
    ):
        # pylint: disable=unused-argument
        i, expected_call = call_queue.take(CALL_TYPE_PUSH_COROSYNC_CONF)

        if not isinstance(corosync_conf_facade, ConfigFacade):
            raise AssertionError(
                (
                    "Trying to call env.push_corosync_conf (call no. {0}) with"
                    " {1} instead of lib.corosync.config_facade.ConfigFacade"
                ).format(i, type(corosync_conf_facade))
            )

        to_push = corosync_conf_facade.config.export()
        if to_push != expected_call.corosync_conf_text:
            raise AssertionError(
                "Trying to call env.push_corosync_conf but the pushed "
                "corosync.conf is not as expected:\n{0}".format(
                    prepare_diff(to_push, expected_call.corosync_conf_text)
                )
            )
        if skip_offline_nodes != expected_call.skip_offline_targets:
            raise AssertionError(
                (
                    "Trying to call env.push_corosync_conf but the "
                    "skip_offline flag is not as expected:\nexpected '{0}' != "
                    "actual '{1}'"
                ).format(
                    expected_call.skip_offline_targets,
                    skip_offline_nodes,
                )
            )
        if (
            corosync_conf_facade.need_stopped_cluster
            != expected_call.need_stopped_cluster
        ):
            raise AssertionError(
                "Tryint to call env.push_corosync_conf but stopped cluster "
                "requirement (corosync_conf_facade.need_stopped_cluster) "
                f"differs. Expected: {expected_call.need_stopped_cluster}; "
                f"Actual: {corosync_conf_facade.need_stopped_cluster}"
            )
        if expected_call.raises:
            raise LibraryError()

    return push_corosync_conf


def is_push_corosync_conf_call_in(call_queue):
    return call_queue.has_type(CALL_TYPE_PUSH_COROSYNC_CONF)



import json

from pcs import utils
from pcs.cli.common.errors import (
    CmdLineInputError,
    ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE,
)
from pcs.cli.common.parse_args import prepare_options
import pcs.lib.pacemaker.live as lib_pacemaker


def node_attribute_cmd(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - allows not unique recipient values
      * --name - specify attribute name to filter out
    """
    del lib
    modifiers.ensure_only_supported("-f", "--force", "--name")
    if modifiers.get("--name") and len(argv) > 1:
        raise CmdLineInputError()
    if not argv:
        attribute_show_cmd(filter_attr=modifiers.get("--name"))
    elif len(argv) == 1:
        attribute_show_cmd(argv.pop(0), filter_attr=modifiers.get("--name"))
    else:
        # --force is used only when setting attributes
        attribute_set_cmd(argv.pop(0), argv)


def node_utilization_cmd(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --name - specify attribute name to filter out
    """
    del lib
    modifiers.ensure_only_supported("-f", "--name")
    if modifiers.get("--name") and len(argv) > 1:
        raise CmdLineInputError()
    if not argv:
        print_node_utilization(filter_name=modifiers.get("--name"))
    elif len(argv) == 1:
        print_node_utilization(argv.pop(0), filter_name=modifiers.get("--name"))
    else:
        set_node_utilization(argv.pop(0), argv)


def node_maintenance_cmd(lib, argv, modifiers, enable):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --wait - wait for node to change state
      * --all - all cluster nodes
    """
    modifiers.ensure_only_supported("-f", "--wait", "--all")
    if argv and modifiers.get("--all"):
        raise CmdLineInputError(ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE)
    wait = modifiers.get("--wait")
    if modifiers.get("--all"):
        lib.node.maintenance_unmaintenance_all(enable, wait)
    elif argv:
        lib.node.maintenance_unmaintenance_list(enable, argv, wait)
    else:
        # -f cannot be used when editing local node
        lib.node.maintenance_unmaintenance_local(enable, wait)


def node_standby_cmd(lib, argv, modifiers, enable):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --wait - wait for node to change state
      * --all - all cluster nodes
    """
    modifiers.ensure_only_supported("-f", "--wait", "--all")
    if argv and modifiers.get("--all"):
        raise CmdLineInputError(ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE)
    wait = modifiers.get("--wait")
    if modifiers.get("--all"):
        lib.node.standby_unstandby_all(enable, wait)
    elif argv:
        lib.node.standby_unstandby_list(enable, argv, wait)
    else:
        # -f cannot be used when editing local node
        lib.node.standby_unstandby_local(enable, wait)


def set_node_utilization(node, argv):
    """
    Commandline options:
      * -f - CIB file
    """
    nvpair_dict = prepare_options(argv)
    if not nvpair_dict:
        return
    only_removing = True
    for value in nvpair_dict.values():
        if value != "":
            only_removing = False
            break

    cib = utils.get_cib_dom()
    node_el = utils.dom_get_node(cib, node)
    if node_el is None:
        if utils.usefile:
            utils.err("Unable to find a node: {0}".format(node))

        for attrs in utils.getNodeAttributesFromPacemaker():
            if attrs.name == node and attrs.type == "remote":
                node_attrs = attrs
                break
        else:
            utils.err("Unable to find a node: {0}".format(node))

        nodes_section_list = cib.getElementsByTagName("nodes")
        if not nodes_section_list:
            utils.err("Unable to get nodes section of cib")

        if only_removing:
            # Do not create new node if we are only removing values from it.
            return

        dom = nodes_section_list[0].ownerDocument
        node_el = dom.createElement("node")
        node_el.setAttribute("id", node_attrs.id)
        node_el.setAttribute("type", node_attrs.type)
        node_el.setAttribute("uname", node_attrs.name)
        nodes_section_list[0].appendChild(node_el)

    utils.dom_update_utilization(node_el, nvpair_dict, "nodes-")
    utils.replace_cib_configuration(cib)


def print_node_utilization(filter_node=None, filter_name=None):
    """
    Commandline options:
      * -f - CIB file
    """
    cib = utils.get_cib_dom()

    node_element_list = cib.getElementsByTagName("node")

    if (
        filter_node
        and filter_node
        not in [
            node_element.getAttribute("uname")
            for node_element in node_element_list
        ]
        and (
            utils.usefile
            or filter_node
            not in [
                node_attrs.name
                for node_attrs in utils.getNodeAttributesFromPacemaker()
            ]
        )
    ):
        utils.err("Unable to find a node: {0}".format(filter_node))

    utilization = {}
    for node_el in node_element_list:
        node = node_el.getAttribute("uname")
        if filter_node is not None and node != filter_node:
            continue
        util_str = utils.get_utilization_str(node_el, filter_name)
        if util_str:
            utilization[node] = util_str
    print("Node Utilization:")
    for node in sorted(utilization):
        print(" {0}: {1}".format(node, utilization[node]))


def node_pacemaker_status(lib, argv, modifiers):
    """
    Internal pcs-pcsd command
    """
    del lib
    del argv
    del modifiers
    print(json.dumps(lib_pacemaker.get_local_node_status(utils.cmd_runner())))


def attribute_show_cmd(filter_node=None, filter_attr=None):
    """
    Commandline options:
      * -f - CIB file (in lib wrapper)
    """
    node_attributes = utils.get_node_attributes(
        filter_node=filter_node, filter_attr=filter_attr
    )
    print("Node Attributes:")
    attribute_print(node_attributes)


def attribute_set_cmd(node, argv):
    """
    Commandline options:
      * -f - CIB file
      * --force - no error if attribute to delete doesn't exist
    """
    for name, value in prepare_options(argv).items():
        utils.set_node_attribute(name, value, node)


def attribute_print(node_attributes):
    """
    Commandline options: no options
    """
    for node in sorted(node_attributes.keys()):
        line_parts = [" " + node + ":"]
        for name, value in sorted(node_attributes[node].items()):
            line_parts.append("{0}={1}".format(name, value))
        print(" ".join(line_parts))



# Reiplementing function pcs.lib.pacemaker.values.is_true to avaoid cyslic
# imports. This is a temporary solution.
def _is_true(val) -> bool:
    return val.lower() in {"true", "on", "yes", "y", "1"}


def constraint_plain(constraint_info, with_id=False):
    """
    dict constraint_info see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    options = constraint_info["options"]
    oc_resource1 = options.get("first", "")
    oc_resource2 = options.get("then", "")
    first_action = options.get("first-action", "")
    then_action = options.get("then-action", "")
    oc_id = options.get("id", "")
    oc_score = options.get("score", "")
    oc_kind = options.get("kind", "")
    oc_sym = ""
    oc_id_out = ""
    oc_options = ""
    if "symmetrical" in options and not _is_true(
        options.get("symmetrical", "false")
    ):
        oc_sym = "(non-symmetrical)"
    if oc_kind != "":
        score_text = "(kind:" + oc_kind + ")"
    elif oc_kind == "" and oc_score == "":
        score_text = "(kind:Mandatory)"
    else:
        score_text = "(score:" + oc_score + ")"
    if with_id:
        oc_id_out = "(id:" + oc_id + ")"
    already_processed_options = (
        "first",
        "then",
        "first-action",
        "then-action",
        "id",
        "score",
        "kind",
        "symmetrical",
    )
    oc_options = " ".join(
        [
            "{0}={1}".format(name, value)
            for name, value in options.items()
            if name not in already_processed_options
        ]
    )
    if oc_options:
        oc_options = "(Options: " + oc_options + ")"
    return " ".join(
        [
            arg
            for arg in [
                first_action,
                oc_resource1,
                "then",
                then_action,
                oc_resource2,
                score_text,
                oc_sym,
                oc_options,
                oc_id_out,
            ]
            if arg
        ]
    )



from unittest import TestCase
from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import (
    get_test_resource as rc,
    get_tmp_file,
    write_data_to_tmpfile,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class EnableDisable(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cib_resource_enable_disable")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_resource(self, name, disabled=False):
        cmd = [
            "resource",
            "create",
            name,
            "ocf:heartbeat:Dummy",
            "--no-default-ops",
        ]
        if disabled:
            cmd.append("--disabled")
        self.assert_pcs_success(cmd)

    def fixture_tag(self, name, ids):
        self.assert_pcs_success(["tag", "create", name] + ids)

    def test_enable_none(self):
        self.assert_pcs_fail(
            "resource enable".split(),
            "Error: You must specify resource(s) to enable\n",
        )

    def test_disable_none(self):
        self.assert_pcs_fail(
            "resource disable".split(),
            "Error: You must specify resource(s) to disable\n",
        )

    def test_enable(self):
        self.fixture_resource("A", disabled=True)
        self.fixture_resource("B", disabled=True)
        self.fixture_tag("TA", ["A"])
        self.assert_effect(
            "resource enable TA B".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes"/>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes"/>
                    <operations>
                        <op id="B-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_disable(self):
        self.fixture_resource("A", disabled=False)
        self.fixture_resource("B", disabled=False)
        self.fixture_tag("TA", ["A"])

        self.assert_effect(
            "resource disable B TA".split(),
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
                <primitive class="ocf" id="B" provider="heartbeat" type="Dummy">
                    <meta_attributes id="B-meta_attributes">
                        <nvpair id="B-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="B-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """,
        )

    def test_enable_nonexistent(self):
        self.fixture_resource("A", disabled=True)

        self.assert_pcs_fail(
            "resource enable A B".split(),
            (
                "Error: bundle/clone/group/resource/tag 'B' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """
        )

    def test_disable_nonexistent(self):
        self.fixture_resource("A", disabled=False)

        self.assert_pcs_fail(
            "resource disable A B".split(),
            (
                "Error: bundle/clone/group/resource/tag 'B' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )
        self.assert_resources_xml_in_cib(
            """
            <resources>
                <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="A-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>
            """
        )

    def test_resource_disable_non_existing_with_failed_action(self):
        xml = """
            <cib epoch="557" num_updates="122" admin_epoch="0" crm_feature_set="3.0.14" validate-with="pacemaker-2.10" update-origin="rh7-3" update-client="crmd" cib-last-written="Thu Aug 23 16:49:17 2012" have-quorum="0" dc-uuid="2">
              <configuration>
                <crm_config/>
                <nodes>
                </nodes>
                <resources/>
                <constraints/>
              </configuration>
              <status>
                <node_state id="1" uname="rh7-1" in_ccm="true" crmd="online" crm-debug-origin="do_update_resource" join="member" expected="member">
                  <lrm id="1">
                    <lrm_resources>
                      <lrm_resource id="A" type="apache" class="ocf" provider="heartbeat">
                        <lrm_rsc_op id="A_last_0" operation_key="A_monitor_0" operation="monitor" crm-debug-origin="do_update_resource" crm_feature_set="3.0.14" transition-key="1:1104:7:817ee250-d179-483a-819e-9be9cb0e06df" transition-magic="0:7;1:1104:7:817ee250-d179-483a-819e-9be9cb0e06df" exit-reason="" on_node="rh7-1" call-id="1079" rc-code="7" op-status="0" interval="0" last-run="1591791198" last-rc-change="1591791198" exec-time="275" queue-time="0" op-digest="f2317cad3d54cec5d7d7aa7d0bf35cf8"/>
                      </lrm_resource>
                    </lrm_resources>
                  </lrm>
                </node_state>
              </status>
            </cib>
            """
        write_data_to_tmpfile(xml, self.temp_cib)
        self.assert_pcs_fail(
            "resource disable A".split(),
            (
                "Error: bundle/clone/group/resource/tag 'A' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )



from pcs_test.tools.assertions import AssertPcsMixin, assert_xml_equal
from pcs_test.tools.misc import write_data_to_tmpfile


def xml_format(xml_string):
    line_list = xml_string.splitlines()
    reindented_lines = [line_list[0]]
    for line in line_list[1:]:
        leading_spaces = len(line) - len(line.lstrip()) - 4
        # current indent is 2 spaces desired is 4 spaces
        indent = " " * 2 * leading_spaces
        new_line = indent + line.strip()
        max_line_len = 80 - 12  # 12 is indent in this file ;)
        if new_line.endswith(">") and len(new_line) > max_line_len:
            last_space = new_line[:max_line_len].rfind(" ")
            if last_space:
                closing = "/>" if new_line.endswith("/>") else ">"
                splited_line = [
                    new_line[:last_space],
                    indent + "   " + new_line[last_space : -1 * len(closing)],
                    indent + closing,
                ]
                reindented_lines.extend(splited_line)
                continue
        # append not splited line
        reindented_lines.append(new_line)

    return "\n".join(reindented_lines)


def get_assert_pcs_effect_mixin(get_cib_part):
    class AssertPcsEffectMixin(AssertPcsMixin):
        def assert_resources_xml_in_cib(
            self,
            expected_xml_resources,
            get_cib_part_func=None,
        ):
            self.temp_cib.seek(0)
            if get_cib_part_func is not None:
                xml = get_cib_part_func(self.temp_cib)
            else:
                xml = get_cib_part(self.temp_cib)
            try:
                assert_xml_equal(expected_xml_resources, xml.decode())
            except AssertionError as e:
                raise AssertionError(
                    "{0}\n\nCopy format ;)\n{1}".format(
                        e.args[0], xml_format(xml.decode())
                    )
                ) from e

        def assert_effect_single(
            self,
            command,
            expected_xml,
            output=None,
            output_start=None,
            output_regexp=None,
        ):
            self.assert_pcs_success(
                command, output, output_start, output_regexp
            )
            self.assert_resources_xml_in_cib(expected_xml)

        def assert_effect(
            self,
            alternative_cmds,
            expected_xml,
            output=None,
            output_start=None,
            output_regexp=None,
        ):
            alternative_list = (
                alternative_cmds
                if isinstance(alternative_cmds[0], list)
                else [alternative_cmds]
            )
            self.temp_cib.seek(0)
            cib_content = self.temp_cib.read()
            self.temp_cib.seek(0)
            for alternative in alternative_list[:-1]:
                self.assert_effect_single(
                    alternative,
                    expected_xml,
                    output,
                    output_start,
                    output_regexp,
                )
                write_data_to_tmpfile(cib_content, self.temp_cib)

            self.assert_effect_single(
                alternative_list[-1],
                expected_xml,
                output,
                output_start,
                output_regexp,
            )

    return AssertPcsEffectMixin



from typing import List

from pcs.common.reports import (
    ReportItem,
    ReportItemSeverity,
    ReportProcessor,
)

from .output import (
    error,
    prepare_force_text,
    warn,
)
from .messages import report_item_msg_from_dto


class ReportProcessorToConsole(ReportProcessor):
    def __init__(self, debug: bool = False) -> None:
        super().__init__()
        self._ignore_severities: List[ReportItemSeverity] = []
        self.debug = debug

    def _do_report(self, report_item: ReportItem) -> None:
        report_dto = report_item.to_dto()
        msg = report_item_msg_from_dto(report_dto.message).message
        if report_dto.context:
            msg = f"{report_dto.context.node}: {msg}"
        severity = report_dto.severity.level

        if severity in self._ignore_severities:
            # DEBUG overrides ignoring severities for debug reports
            if msg and self.debug and severity == ReportItemSeverity.DEBUG:
                print(msg)
            return

        if severity == ReportItemSeverity.ERROR:
            error(
                "{msg}{force}".format(
                    msg=msg,
                    force=prepare_force_text(report_item),
                )
            )
        elif severity == ReportItemSeverity.WARNING:
            warn(msg)
        elif msg and (self.debug or severity != ReportItemSeverity.DEBUG):
            print(msg)

    def suppress_reports_of_severity(
        self, severity_list: List[ReportItemSeverity]
    ) -> None:
        self._ignore_severities = list(severity_list)



from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource as rc,
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

from pcs.common import const
from pcs.common.str_tools import format_list


class ParseArgvDashDash(TestCase, AssertPcsMixin):
    # The command will fail, that's ok. We are interested only in the argv
    # parsing.
    cmd = "constraint colocation add R1 with R2".split()

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_misc")
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.allowed_roles = format_list(const.PCMK_ROLES)

    def tearDown(self):
        self.temp_cib.close()

    def test_negative_int(self):
        self.assert_pcs_fail(
            self.cmd + ["-123"],
            outdent(
                """\
                Error: Resource 'R1' does not exist
                Warning: Using '-123' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                """
            ),
        )

    def test_negative_float(self):
        self.assert_pcs_fail(
            self.cmd + ["-12.3"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                Warning: Using '-12.3' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                """
            ),
        )

    def test_negative_infinity(self):
        self.assert_pcs_fail(
            self.cmd + ["-inFIniTY"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                Warning: Using '-inFIniTY' without '--' is deprecated, those parameters will be considered position independent options in future pcs versions
                """
            ),
        )

    def test_negative_int_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-123"],
            outdent(
                """\
                Error: Resource 'R1' does not exist
                """
            ),
        )

    def test_negative_float_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-12.3"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )

    def test_negative_infinity_dash_dash(self):
        self.assert_pcs_fail(
            ["--"] + self.cmd + ["-inFIniTY"],
            outdent(
                f"""\
                Error: invalid role value 'R2', allowed values are: {self.allowed_roles}
                """
            ),
        )



from unittest import TestCase
from contextlib import contextmanager

from pcs_test.tools.misc import create_setup_patch_mixin

from pcs.daemon import session
from pcs.daemon.session import Session

SID = "abc"
USER = "user"
GROUPS = ["group1", "group2"]


class AssertMixin:
    def assert_vanila_session(self, _session):
        self.assertIsNone(_session.username)
        self.assertFalse(_session.is_authenticated)
        self.assertIsNone(_session.ajax_id)
        self.assertEqual(_session.groups, [])

    def assert_authenticated_session(self, _session, username, groups):
        self.assertEqual(_session.username, username)
        self.assertEqual(_session.groups, groups)
        self.assertTrue(_session.is_authenticated)
        self.assertTrue(_session.ajax_id is not None)

    def assert_login_failed_session(self, _session, username):
        self.assertEqual(_session.username, username)
        self.assertFalse(_session.is_authenticated)


PatchSessionMixin = create_setup_patch_mixin(session)


class SessionTest(TestCase, AssertMixin, PatchSessionMixin):
    def setUp(self):
        self.now = self.setup_patch("now", return_value=0)
        self.session = Session(SID)

    def test_session_grows_older(self):
        self.now.return_value = 10.1
        self.assertTrue(self.session.was_unused_last(10))
        self.assertFalse(self.session.was_unused_last(11))

    @contextmanager
    def refresh_test(self):
        self.now.return_value = 10.1
        yield self.session
        self.now.return_value = 11.2
        self.assertTrue(self.session.was_unused_last(1))
        self.assertFalse(self.session.was_unused_last(2))

    def test_session_is_refreshable(self):
        # pylint: disable=pointless-statement
        with self.refresh_test() as session1:
            session1.refresh()
        with self.refresh_test() as session1:
            session1.is_authenticated
        with self.refresh_test() as session1:
            session1.username
        with self.refresh_test() as session1:
            session1.groups
        with self.refresh_test() as session1:
            session1.sid
        with self.refresh_test() as session1:
            session1.ajax_id


class StorageTest(TestCase, AssertMixin, PatchSessionMixin):
    def setUp(self):
        self.now = self.setup_patch("now", return_value=0)
        self.storage = session.Storage(lifetime_seconds=10)

    def test_creates_vanilla_session_when_sid_not_specified(self):
        self.assert_vanila_session(self.storage.provide())

    def test_does_not_accept_foreign_sid(self):
        session1 = self.storage.provide("unknown_sid")
        self.assertNotEqual(session1.sid, "unknown_sid")
        self.assert_vanila_session(session1)

    def test_provides_the_same_session_for_same_sid(self):
        session1 = self.storage.provide()
        session2 = self.storage.provide(session1.sid)
        self.assertIs(session1, session2)

    def test_can_destroy_session(self):
        session1 = self.storage.provide()
        self.storage.destroy(session1.sid)
        session2 = self.storage.provide(session1.sid)
        self.assertIsNot(session1, session2)

    def test_can_drop_expired_sessions_explicitly(self):
        session1 = self.storage.provide()
        self.now.return_value = 5
        session2 = self.storage.provide()
        self.now.return_value = 12
        self.storage.drop_expired()
        session3 = self.storage.provide(session1.sid)
        session4 = self.storage.provide(session2.sid)
        self.assertIsNot(session3, session1)
        self.assertIs(session4, session2)

    def test_can_drop_expired_session_implicitly(self):
        session1 = self.storage.provide()
        sid = session1.sid
        self.now.return_value = 11
        session2 = self.storage.provide(sid)
        self.assertIsNot(session1, session2)

    def test_can_login_new_session(self):
        self.assert_authenticated_session(
            self.storage.login(sid=None, username=USER, groups=GROUPS),
            USER,
            GROUPS,
        )

    def test_can_login_existing_session(self):
        session1 = self.storage.provide()
        session2 = self.storage.login(session1.sid, USER, GROUPS)
        self.assert_authenticated_session(session2, USER, GROUPS)
        self.assertEqual(session1.sid, session2.sid)

    def test_can_sign_failed_login_attempt_new_session(self):
        self.assert_login_failed_session(
            self.storage.rejected_user(sid=None, username=USER), USER
        )

    def test_can_sign_failed_login_attempt_existing_session(self):
        session1 = self.storage.provide()
        session2 = self.storage.rejected_user(session1.sid, USER)
        self.assert_login_failed_session(session2, USER)
        self.assertEqual(session1.sid, session2.sid)



from textwrap import dedent
from unittest import mock, TestCase

from pcs_test.tools.misc import dict_to_modifiers

from pcs.common.reports import codes as report_codes

from pcs.cli import dr
from pcs.cli.common.errors import CmdLineInputError


@mock.patch("pcs.cli.dr.print")
class Config(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["dr"])
        self.lib.dr = mock.Mock(spec_set=["get_config"])

    def _call_cmd(self, argv=None):
        dr.config(self.lib, argv or [], dict_to_modifiers({}))

    def test_argv(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["x"])
        self.assertIsNone(cm.exception.message)
        mock_print.assert_not_called()

    def test_success(self, mock_print):
        self.lib.dr.get_config.return_value = {
            "local_site": {
                "node_list": [],
                "site_role": "RECOVERY",
            },
            "remote_site_list": [
                {
                    "node_list": [
                        {"name": "nodeA2"},
                        {"name": "nodeA1"},
                    ],
                    "site_role": "PRIMARY",
                },
                {
                    "node_list": [
                        {"name": "nodeB1"},
                    ],
                    "site_role": "RECOVERY",
                },
            ],
        }
        self._call_cmd([])
        self.lib.dr.get_config.assert_called_once_with()
        mock_print.assert_called_once_with(
            dedent(
                """\
            Local site:
              Role: Recovery
            Remote site:
              Role: Primary
              Nodes:
                nodeA1
                nodeA2
            Remote site:
              Role: Recovery
              Nodes:
                nodeB1"""
            )
        )

    @mock.patch("pcs.cli.reports.output.sys.stderr.write")
    def test_invalid_response(self, mock_stderr, mock_print):
        self.lib.dr.get_config.return_value = [
            "wrong response",
            {"x": "y"},
        ]
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd([])
        self.assertEqual(cm.exception.code, 1)
        self.lib.dr.get_config.assert_called_once_with()
        mock_print.assert_not_called()
        mock_stderr.assert_called_once_with(
            "Error: Unable to communicate with pcsd, received response:\n"
            "['wrong response', {'x': 'y'}]\n"
        )


class SetRecoverySite(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["dr"])
        self.dr = mock.Mock(spec_set=["set_recovery_site"])
        self.lib.dr = self.dr

    def call_cmd(self, argv):
        dr.set_recovery_site(self.lib, argv, dict_to_modifiers({}))

    def test_no_node(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_multiple_nodes(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["node1", "node2"])
        self.assertIsNone(cm.exception.message)

    def test_success(self):
        node = "node"
        self.call_cmd([node])
        self.dr.set_recovery_site.assert_called_once_with(node)


@mock.patch("pcs.cli.dr.print")
class Status(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["dr"])
        self.lib.dr = mock.Mock(spec_set=["status_all_sites_plaintext"])

    def _call_cmd(self, argv, modifiers=None):
        dr.status(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def _fixture_response(self, local_success=True, remote_success=True):
        self.lib.dr.status_all_sites_plaintext.return_value = [
            {
                "local_site": True,
                "site_role": "PRIMARY",
                "status_plaintext": (
                    "local cluster\nstatus"
                    if local_success
                    else "this should never be displayed"
                ),
                "status_successfully_obtained": local_success,
            },
            {
                "local_site": False,
                "site_role": "RECOVERY",
                "status_plaintext": (
                    "remote cluster\nstatus"
                    if remote_success
                    else "this should never be displayed"
                ),
                "status_successfully_obtained": remote_success,
            },
        ]

    @staticmethod
    def _fixture_print():
        return dedent(
            """\
            --- Local cluster - Primary site ---
            local cluster
            status


            --- Remote cluster - Recovery site ---
            remote cluster
            status"""
        )

    def test_argv(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["x"])
        self.assertIsNone(cm.exception.message)
        mock_print.assert_not_called()

    def test_success(self, mock_print):
        self._fixture_response()
        self._call_cmd([])
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=False
        )
        mock_print.assert_called_once_with(self._fixture_print())

    def test_success_full(self, mock_print):
        self._fixture_response()
        self._call_cmd([], {"full": True})
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=True
        )
        mock_print.assert_called_once_with(self._fixture_print())

    def test_success_hide_inactive(self, mock_print):
        self._fixture_response()
        self._call_cmd([], {"hide-inactive": True})
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=True, verbose=False
        )
        mock_print.assert_called_once_with(self._fixture_print())

    def test_success_all_flags(self, mock_print):
        self._fixture_response()
        self._call_cmd([], {"full": True, "hide-inactive": True})
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=True, verbose=True
        )
        mock_print.assert_called_once_with(self._fixture_print())

    @mock.patch("pcs.cli.reports.output.sys.stderr.write")
    def test_error_local(self, mock_stderr, mock_print):
        self._fixture_response(local_success=False)
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd([])
        self.assertEqual(cm.exception.code, 1)
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=False
        )
        mock_print.assert_called_once_with(
            dedent(
                """\
            --- Local cluster - Primary site ---
            Error: Unable to get status of the cluster from any node

            --- Remote cluster - Recovery site ---
            remote cluster
            status"""
            )
        )
        mock_stderr.assert_called_once_with(
            "Error: Unable to get status of all sites\n"
        )

    @mock.patch("pcs.cli.reports.output.sys.stderr.write")
    def test_error_remote(self, mock_stderr, mock_print):
        self._fixture_response(remote_success=False)
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd([])
        self.assertEqual(cm.exception.code, 1)
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=False
        )
        mock_print.assert_called_once_with(
            dedent(
                """\
            --- Local cluster - Primary site ---
            local cluster
            status


            --- Remote cluster - Recovery site ---
            Error: Unable to get status of the cluster from any node"""
            )
        )
        mock_stderr.assert_called_once_with(
            "Error: Unable to get status of all sites\n"
        )

    @mock.patch("pcs.cli.reports.output.sys.stderr.write")
    def test_error_both(self, mock_stderr, mock_print):
        self._fixture_response(local_success=False, remote_success=False)
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd([])
        self.assertEqual(cm.exception.code, 1)
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=False
        )
        mock_print.assert_called_once_with(
            dedent(
                """\
            --- Local cluster - Primary site ---
            Error: Unable to get status of the cluster from any node

            --- Remote cluster - Recovery site ---
            Error: Unable to get status of the cluster from any node"""
            )
        )
        mock_stderr.assert_called_once_with(
            "Error: Unable to get status of all sites\n"
        )

    @mock.patch("pcs.cli.reports.output.sys.stderr.write")
    def test_invalid_response(self, mock_stderr, mock_print):
        self.lib.dr.status_all_sites_plaintext.return_value = [
            "wrong response",
            {"x": "y"},
        ]
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd([])
        self.assertEqual(cm.exception.code, 1)
        self.lib.dr.status_all_sites_plaintext.assert_called_once_with(
            hide_inactive_resources=False, verbose=False
        )
        mock_print.assert_not_called()
        mock_stderr.assert_called_once_with(
            "Error: Unable to communicate with pcsd, received response:\n"
            "['wrong response', {'x': 'y'}]\n"
        )


class Destroy(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["dr"])
        self.dr = mock.Mock(spec_set=["destroy"])
        self.lib.dr = self.dr

    def call_cmd(self, argv, modifiers=None):
        modifiers = modifiers or {}
        dr.destroy(self.lib, argv, dict_to_modifiers(modifiers))

    def test_some_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["arg"])
        self.assertIsNone(cm.exception.message)

    def test_success(self):
        self.call_cmd([])
        self.dr.destroy.assert_called_once_with(force_flags=[])

    def test_skip_offline(self):
        self.call_cmd([], modifiers={"skip-offline": True})
        self.dr.destroy.assert_called_once_with(
            force_flags=[report_codes.SKIP_OFFLINE_NODES]
        )



from functools import partial
from unittest import mock, TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from pcs import settings
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import stonith
from pcs.lib.resource_agent import StonithAgent

expected_cib_simple = """
    <primitive class="stonith" id="stonith-test" type="test_simple">
        <instance_attributes id="stonith-test-instance_attributes">
            <nvpair id="stonith-test-instance_attributes-must-set"
                name="must-set" value="value"
            />
            <nvpair id="stonith-test-instance_attributes-must-set-new"
                name="must-set-new" value="B"
            />
        </instance_attributes>
        <operations>
            <op id="stonith-test-monitor-interval-60s" interval="60s"
                name="monitor"
            />
        </operations>
    </primitive>
"""

expected_cib_simple_forced = """
    <primitive class="stonith" id="stonith-test" type="test_simple">
        <instance_attributes id="stonith-test-instance_attributes">
            <nvpair id="stonith-test-instance_attributes-undefined"
                name="undefined" value="attribute"
            />
        </instance_attributes>
        <meta_attributes id="stonith-test-meta_attributes">
            <nvpair id="stonith-test-meta_attributes-metaname"
                name="metaname" value="metavalue"
            />
        </meta_attributes>
        <operations>
            <op id="stonith-test-bad-action-interval-0s" interval="0s"
                name="bad-action"
            />
            <op id="stonith-test-monitor-interval-60s" interval="60s"
                name="monitor"
            />
        </operations>
    </primitive>
"""

expected_cib_unfencing = """
    <primitive class="stonith" id="stonith-test" type="test_unfencing">
        <meta_attributes id="stonith-test-meta_attributes">
            <nvpair id="stonith-test-meta_attributes-provides"
                name="provides" value="unfencing"
            />
        </meta_attributes>
        <operations>
            <op id="stonith-test-monitor-interval-60s"
                interval="60s" name="monitor"
            />
        </operations>
    </primitive>
"""

expected_cib_operations = """
    <primitive class="stonith" id="stonith-test" type="test_custom_actions">
        <operations>
            <op id="stonith-test-monitor-interval-27s"
                interval="27s" name="monitor" timeout="11s"
            />
        </operations>
    </primitive>
"""

expected_cib_unknown = """
    <primitive class="stonith" id="stonith-test" type="test_unknown">
        <operations>
            <op id="stonith-test-monitor-interval-60s" interval="60s"
                name="monitor"
            />
        </operations>
    </primitive>
"""


class CreateMixin:
    def setUp(self):
        # pylint does not know this method is defined in TestCase
        # pylint: disable=invalid-name
        self.env_assist, self.config = get_env_tools(test_case=self)

    def tearDown(self):
        # pylint does not know this method is defined in TestCase
        # pylint: disable=invalid-name
        # pylint: disable=no-self-use
        StonithAgent.clear_fenced_metadata_cache()

    def test_minimal_success(self):
        agent_name = "test_simple"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(resources=self._expected_cib(expected_cib_simple))
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={
                "must-set": "value",
                "must-set-new": "B",
            },
        )

    def test_unfencing(self):
        agent_name = "test_unfencing"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_unfencing.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(resources=self._expected_cib(expected_cib_unfencing))
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={},
        )

    def test_disabled(self):
        agent_name = "test_simple"
        expected_cib = expected_cib_simple.replace(
            '<instance_attributes id="stonith-test-instance_attributes">',
            """
                <meta_attributes id="stonith-test-meta_attributes">
                    <nvpair id="stonith-test-meta_attributes-target-role"
                        name="target-role" value="Stopped"
                    />
                </meta_attributes>
                <instance_attributes id="stonith-test-instance_attributes">
            """,
        )

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(resources=self._expected_cib(expected_cib))
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={
                "must-set": "value",
                "must-set-new": "B",
            },
            ensure_disabled=True,
        )

    def _assert_default_operations(self, use_default_operations):
        # use_default_operations currently has no effect because in both cases
        # only the monitor operation is created in cib. That is correct. Still
        # it is worth testing. If it ever changes, the test should fail and be
        # updated to test new behaviour.
        agent_name = "test_custom_actions"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_custom_actions.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(resources=self._expected_cib(expected_cib_operations))
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={},
            use_default_operations=use_default_operations,
        )

    def test_default_operations_yes(self):
        self._assert_default_operations(True)

    def test_default_operations_no(self):
        self._assert_default_operations(False)

    def test_id_already_exists(self):
        agent_name = "test_simple"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            ).runner.cib.load(resources=self._expected_cib(expected_cib_simple))
        )

        self.env_assist.assert_raise_library_error(
            lambda: self._create(
                self.env_assist.get_env(),
                "stonith-test",
                agent_name,
                operations=[],
                meta_attributes={},
                instance_attributes={
                    "must-set": "value",
                    "must-set-new": "B",
                },
            ),
            [fixture.error(report_codes.ID_ALREADY_EXISTS, id="stonith-test")],
            expected_in_processor=False,
        )

    def test_instance_meta_and_operations(self):
        agent_name = "test_simple"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(
                resources=self._expected_cib(expected_cib_simple_forced)
            )
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[
                {"name": "bad-action"},
            ],
            meta_attributes={
                "metaname": "metavalue",
            },
            instance_attributes={"undefined": "attribute"},
            allow_invalid_operation=True,
            allow_invalid_instance_attributes=True,
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="bad-action",
                    option_name="operation name",
                    allowed_values=[
                        "on",
                        "off",
                        "reboot",
                        "status",
                        "list",
                        "list-status",
                        "monitor",
                        "metadata",
                        "validate-all",
                    ],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.warn(
                    report_codes.INVALID_OPTIONS,
                    option_names=["undefined"],
                    option_type="stonith",
                    allowed=[
                        "may-set",
                        "must-set",
                        "must-set-new",
                        "must-set-old",
                        "pcmk_action_limit",
                        "pcmk_delay_base",
                        "pcmk_delay_max",
                        "pcmk_host_argument",
                        "pcmk_host_check",
                        "pcmk_host_list",
                        "pcmk_host_map",
                        "pcmk_list_action",
                        "pcmk_list_retries",
                        "pcmk_list_timeout",
                        "pcmk_monitor_action",
                        "pcmk_monitor_retries",
                        "pcmk_monitor_timeout",
                        "pcmk_off_action",
                        "pcmk_off_retries",
                        "pcmk_off_timeout",
                        "pcmk_on_action",
                        "pcmk_on_retries",
                        "pcmk_on_timeout",
                        "pcmk_reboot_action",
                        "pcmk_reboot_retries",
                        "pcmk_reboot_timeout",
                        "pcmk_status_action",
                        "pcmk_status_retries",
                        "pcmk_status_timeout",
                        "priority",
                    ],
                    allowed_patterns=[],
                ),
                fixture.warn(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["must-set", "must-set-new"],
                    option_type="stonith",
                ),
            ]
        )

    def test_unknown_agent_forced(self):
        agent_name = "test_unknown"
        self.config.runner.pcmk.load_agent(
            agent_name=f"stonith:{agent_name}",
            agent_is_missing=True,
        )
        self.config.runner.cib.load()
        self.config.env.push_cib(
            resources=self._expected_cib(expected_cib_unknown)
        )

        self._create(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={},
            allow_absent_agent=True,
        )

        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="test_unknown",
                    reason=(
                        "Agent stonith:test_unknown not found or does not "
                        "support meta-data: Invalid argument (22)\n"
                        "Metadata query for stonith:test_unknown failed: "
                        "Input/output error"
                    ),
                ),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_minimal_wait_ok_run_ok(self):
        agent_name = "test_simple"
        instance_name = "stonith-test"
        timeout = 10
        expected_status = """
            <resources>
                <resource
                    id="{id}"
                    resource_agent="stonith:{agent}"
                    role="Started"
                    active="true"
                    failed="false"
                    nodes_running_on="1"
                >
                    <node name="node1" id="1" cached="false"/>
                </resource>
            </resources>
            """.format(
            id=instance_name, agent=agent_name
        )

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(
                resources=self._expected_cib(expected_cib_simple), wait=timeout
            )
            .runner.pcmk.load_state(resources=expected_status)
        )

        self._create(
            self.env_assist.get_env(),
            instance_name,
            agent_name,
            operations=[],
            meta_attributes={},
            instance_attributes={
                "must-set": "value",
                "must-set-new": "B",
            },
            wait=str(timeout),
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id=instance_name,
                ),
            ]
        )


class Create(CreateMixin, TestCase):
    _create = staticmethod(stonith.create)

    @staticmethod
    def _expected_cib(xml):
        return "<resources>" + xml + "</resources>"


class CreateInGroup(CreateMixin, TestCase):
    _create = staticmethod(
        partial(stonith.create_in_group, group_id="my-group")
    )

    @staticmethod
    def _expected_cib(xml):
        return "<resources><group id='my-group'>" + xml + "</group></resources>"

    @staticmethod
    def _dummy(name):
        return f"""
            <primitive class="ocf" id="{name}" provider="pacemaker" type="Dummy"
            />
        """

    def test_group_not_valid(self):
        agent_name = "test_simple"

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load()
            .runner.pcmk.load_fenced_metadata()
        )

        self.env_assist.assert_raise_library_error(
            lambda: stonith.create_in_group(
                self.env_assist.get_env(),
                "stonith-test",
                agent_name,
                "0-group",
                operations=[],
                meta_attributes={},
                instance_attributes={
                    "must-set": "value",
                    "must-set-new": "B",
                },
            ),
            [
                fixture.error(
                    report_codes.INVALID_ID_BAD_CHAR,
                    id="0-group",
                    id_description="group name",
                    is_first_char=True,
                    invalid_character="0",
                )
            ],
            expected_in_processor=False,
        )

    def _assert_adjacent(self, adjacent, after):
        agent_name = "test_simple"
        original_cib = (
            "<resources><group id='my-group'>"
            + self._dummy("dummy1")
            + self._dummy("dummy2")
            + "</group></resources>"
        )
        expected_cib = (
            "<resources><group id='my-group'>"
            + self._dummy("dummy1")
            + expected_cib_simple
            + self._dummy("dummy2")
            + "</group></resources>"
        )

        (
            self.config.runner.pcmk.load_agent(
                agent_name=f"stonith:{agent_name}",
                agent_filename="stonith_agent_fence_simple.xml",
            )
            .runner.cib.load(resources=original_cib)
            .runner.pcmk.load_fenced_metadata()
            .env.push_cib(resources=expected_cib)
        )

        stonith.create_in_group(
            self.env_assist.get_env(),
            "stonith-test",
            agent_name,
            "my-group",
            operations=[],
            meta_attributes={},
            instance_attributes={
                "must-set": "value",
                "must-set-new": "B",
            },
            adjacent_resource_id=adjacent,
            put_after_adjacent=after,
        )

    def test_put_after_adjacent(self):
        self._assert_adjacent("dummy1", True)

    def test_put_before_adjacent(self):
        self._assert_adjacent("dummy2", False)
